from __future__ import annotations

import argparse
import json
import sys
from typing import Any

from kognic.auth.cli import _configure_logging
from kognic.auth.cli._output import _print_response
from kognic.auth.env_config import DEFAULT_ENV_CONFIG_FILE_PATH, load_kognic_env_config, resolve_environment
from kognic.auth.internal.token_cache import make_cache
from kognic.auth.requests.base_client import create_session, make_token_provider

METHODS = ["get", "post", "put", "patch", "delete", "head", "options"]


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="kog",
        description="Make authenticated HTTP requests to Kognic APIs",
    )
    parser.add_argument("method", metavar="METHOD", choices=METHODS, help=f"HTTP method ({', '.join(METHODS)})")
    parser.add_argument("url", metavar="URL", help="Full URL to call")
    parser.add_argument("-d", "--data", help="Request body (JSON string)")
    parser.add_argument(
        "-H",
        "--header",
        action="append",
        dest="headers",
        metavar="HEADER",
        help="Header in 'Key: Value' format (repeatable)",
    )
    parser.add_argument(
        "--env-config-file-path",
        default=DEFAULT_ENV_CONFIG_FILE_PATH,
        help=f"Environment config file path (default: {DEFAULT_ENV_CONFIG_FILE_PATH})",
    )
    parser.add_argument("--env", dest="env_name", help="Force a specific environment (skip URL-based matching)")
    parser.add_argument(
        "--format",
        dest="output_format",
        choices=["json", "jsonl", "csv", "tsv", "table"],
        default="json",
        help="Output format: json (default), jsonl (one JSON object per line), csv, tsv, table (markdown)",
    )
    parser.add_argument(
        "--token-cache",
        choices=["auto", "keyring", "file", "none"],
        default="auto",
        help="Token cache backend: auto (default), keyring, file, or none. "
        "Auto will use keyring if available, otherwise file.",
    )
    parser.add_argument("-v", "--verbose", action="store_true", default=False, help="Enable debug logging")
    return parser


def _parse_headers(raw: list[str] | None) -> dict[str, str] | None:
    if not raw:
        return None
    headers: dict[str, str] = {}
    for h in raw:
        if ": " not in h:
            raise ValueError(f"Invalid header format '{h}'. Expected 'Key: Value'.")
        key, value = h.split(": ", 1)
        headers[key] = value
    return headers


def _parse_body(raw: str | None, headers: dict[str, str]) -> Any:
    if not raw:
        return None
    try:
        data = json.loads(raw)
    except json.JSONDecodeError as e:
        raise ValueError(f"Invalid JSON data: {e}") from e
    headers.setdefault("Content-Type", "application/json")
    return data


def run(parsed: argparse.Namespace) -> int:
    try:
        headers = _parse_headers(parsed.headers) or {}
        data = _parse_body(parsed.data, headers)

        config = load_kognic_env_config(parsed.env_config_file_path)
        env = resolve_environment(config, parsed.url, parsed.env_name)

        provider = make_token_provider(
            auth=env.credentials, auth_host=env.auth_server, token_cache=make_cache(parsed.token_cache)
        )
        session = create_session(token_provider=provider)

        response = session.request(
            method=parsed.method.upper(),
            url=parsed.url,
            json=data if data is not None else None,
            headers=headers if headers else None,
        )

        _print_response(response, output_format=parsed.output_format)
        return 0 if response.ok else 1

    except (FileNotFoundError, ValueError, json.JSONDecodeError) as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1


def main(args: list[str] | None = None) -> None:
    parser = _create_parser()
    parsed = parser.parse_args(args)
    _configure_logging(verbose=parsed.verbose)
    sys.exit(run(parsed))



from __future__ import annotations

import argparse
import json
import sys

from kognic.auth.credentials_parser import parse_credentials
from kognic.auth.internal.credentials_store import (
    DEFAULT_PROFILE,
    clear_credentials,
    load_credentials,
    save_credentials,
)

COMMAND = "credentials"


def register_parser(subparsers: argparse._SubParsersAction) -> argparse.ArgumentParser:
    parser = subparsers.add_parser(
        COMMAND,
        help="Manage stored credentials in the system keyring",
    )
    subs = parser.add_subparsers(dest="credentials_action")

    put_p = subs.add_parser("put", help="Put credentials from a JSON file into the system keyring")
    put_p.add_argument("file", metavar="FILE", help="Path to credentials JSON file")
    put_p.add_argument(
        "--env",
        default=DEFAULT_PROFILE,
        metavar="ENV",
        help=f"Keyring profile name to store credentials under (default: {DEFAULT_PROFILE}). "
        "Use the environment name from environments.json to link credentials to that environment "
        "(e.g. --env production → use 'keyring://production' in your config).",
    )

    get_p = subs.add_parser("get", help="Get stored credentials from the system keyring")
    get_p.add_argument(
        "--env",
        default=DEFAULT_PROFILE,
        metavar="ENV",
        help=f"Keyring profile name to read credentials from (default: {DEFAULT_PROFILE}).",
    )

    clear_p = subs.add_parser("delete", help="Delete stored credentials from the system keyring")
    clear_p.add_argument(
        "--env",
        default=DEFAULT_PROFILE,
        metavar="ENV",
        help=f"Keyring profile name to clear (default: {DEFAULT_PROFILE}).",
    )

    return parser


def run(parsed: argparse.Namespace) -> int:
    if parsed.credentials_action == "put":
        return _run_put(parsed)
    if parsed.credentials_action == "get":
        return _run_get(parsed)
    if parsed.credentials_action == "delete":
        return _run_delete(parsed)
    return 0


def _run_put(parsed: argparse.Namespace) -> int:
    try:
        creds = parse_credentials(parsed.file)
        save_credentials(creds, parsed.env)
        print(f"Credentials for client_id={creds.client_id!r} stored in keyring (profile={parsed.env!r})")
        return 0
    except FileNotFoundError as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1
    except (KeyError, ValueError, RuntimeError) as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1


def _run_get(parsed: argparse.Namespace) -> int:
    try:
        creds = load_credentials(parsed.env)
        if creds is None:
            print(f"No credentials found in keyring (profile={parsed.env!r})", file=sys.stderr)
            return 1
        data = {
            "clientId": creds.client_id,
            "clientSecret": creds.client_secret,
            "email": creds.email,
            "userId": creds.user_id,
            "issuer": creds.issuer,
        }
        print(json.dumps(data, indent=2))
        return 0
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1


def _run_delete(parsed: argparse.Namespace) -> int:
    try:
        clear_credentials(parsed.env)
        print(f"Credentials cleared from keyring (profile={parsed.env!r})")
        return 0
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1



"""Base API client V2 using requests/OAuth2 session."""

from __future__ import annotations

import logging
import os
import threading
from threading import Lock
from typing import TYPE_CHECKING, Any, Callable, Optional, Union
from weakref import WeakValueDictionary

if TYPE_CHECKING:
    from typing import Self

import requests
from requests import Session
from requests.adapters import HTTPAdapter, Retry

from kognic.auth import DEFAULT_HOST, DEFAULT_TOKEN_ENDPOINT_RELPATH
from kognic.auth._sunset import SunsetHandler, default_sunset_handler, handle_sunset
from kognic.auth._user_agent import get_user_agent
from kognic.auth.credentials_parser import ANY_AUTH_TYPE, resolve_credentials
from kognic.auth.env_config import DEFAULT_ENV_CONFIG_FILE_PATH, load_kognic_env_config
from kognic.auth.internal.token_cache import TokenCache
from kognic.auth.requests.auth_session import RequestsAuthSession
from kognic.auth.requests.bearer_auth import KognicBearerAuth
from kognic.auth.serde import serialize_body

logger = logging.getLogger(__name__)


DEFAULT_RETRY = Retry(total=3, connect=3, read=3, backoff_factor=0.5, status_forcelist=[502, 503, 504])

_DEFAULT_SUNSET_HANDLER: SunsetHandler = default_sunset_handler()


def _check_response(resp: requests.Response, sunset_handler: Optional[SunsetHandler] = _DEFAULT_SUNSET_HANDLER):
    """Handle sunset headers and raise for status with enhanced error messages."""
    handle_sunset(resp, sunset_handler)
    try:
        resp.raise_for_status()
    except requests.HTTPError as e:
        try:
            js = resp.json()
            err_message = js.get("message", js)
        except ValueError:
            err_message = resp.text
        raise requests.HTTPError(
            f"Got HttpError with status={resp.status_code} in call to {resp.url}.\n"
            f"Got error in response: '{err_message}'",
            response=resp,
            request=resp.request,
        ) from e


def _set_session_user_agent(session: Session, client_name: Optional[str] = None):
    """Set the User-Agent header for the session, including the client name if provided."""
    session.headers["User-Agent"] = get_user_agent(f"requests/{requests.__version__}", client_name)


def _monkey_patch_send(
    session: Session,
    json_serializer: Callable[[Any], Any],
    sunset_handler: Optional[SunsetHandler] = _DEFAULT_SUNSET_HANDLER,
):
    """
    Monkey patch to serialize JSON and validate paths
    :param session:
    :param json_serializer:
    :return:
    """
    vanilla_prep = session.prepare_request

    def prepare_request(req, *args, **kwargs):
        if req.url.startswith("/"):
            raise ValueError(f"Path must not start with /, got {req.url}")

        # Accept anything jsonable as json, serialize it
        if req.json is not None:
            req.json = json_serializer(req.json)

        return vanilla_prep(req, *args, **kwargs)

    session.prepare_request = prepare_request

    # Monkey patch to always raise for status and handle errors
    vanilla_send = session.send

    def send_request(req, *args, **kwargs):
        resp = vanilla_send(req, *args, **kwargs)
        _check_response(resp, sunset_handler)
        return resp

    session.send = send_request


def create_session(
    *,
    auth: Optional[Union[str, os.PathLike, tuple]] = None,
    auth_host: str = DEFAULT_HOST,
    auth_token_endpoint: str = DEFAULT_TOKEN_ENDPOINT_RELPATH,
    client_name: Optional[str] = None,
    json_serializer: Callable[[Any], Any] = serialize_body,
    initial_token: Optional[dict] = None,
    on_token_updated: Optional[Callable[[dict], None]] = None,
    token_provider: Optional[RequestsAuthSession] = None,
    sunset_handler: Optional[SunsetHandler] = _DEFAULT_SUNSET_HANDLER,
) -> Session:
    """Create a requests session with enhancements.

    - OAuth2 authentication with automatic token refresh
    - Automatic JSON serialization for jsonable objects
    - Default retry logic for transient errors
    - Sunset header handling
    - Always call raise_for_status with enhanced error messages

    Args:
        auth: Authentication credentials - path to credentials file or (client_id, client_secret) tuple
        auth_host: Authentication server base URL
        auth_token_endpoint: Relative path to token endpoint
        client_name: Name added to User-Agent header
        json_serializer: Callable to serialize request bodies. Defaults to serialize_body.
        initial_token: Pre-fetched token dict to inject, skipping the initial network fetch if valid.
        on_token_updated: Callback invoked with the new token dict whenever a fresh token is fetched.
        token_provider: Explicit token provider to use. When given, auth/initial_token/on_token_updated
            are ignored. Multiple sessions sharing one provider share the same token lifecycle.
        sunset_handler: Callable invoked with ``(sunset_date, method, url)`` when a sunset header is
            detected. Defaults to logging a warning or error. Pass ``None`` to disable.

    Returns:
        Configured requests Session
    """
    if token_provider is None:
        token_provider = RequestsAuthSession(
            auth=auth,
            host=auth_host,
            token_endpoint=auth_token_endpoint,
            initial_token=initial_token,
            on_token_updated=on_token_updated,
        )

    session = requests.Session()
    session.auth = KognicBearerAuth(token_provider)
    _set_session_user_agent(session, client_name)
    _monkey_patch_send(session, json_serializer, sunset_handler)
    session.mount("http://", HTTPAdapter(max_retries=DEFAULT_RETRY))
    session.mount("https://", HTTPAdapter(max_retries=DEFAULT_RETRY))
    return session


def make_token_provider(
    *,
    auth: ANY_AUTH_TYPE = None,
    auth_host: str = DEFAULT_HOST,
    auth_token_endpoint: str = DEFAULT_TOKEN_ENDPOINT_RELPATH,
    token_cache: Optional[TokenCache] = None,
) -> RequestsAuthSession:
    """Create a RequestsAuthSession wired to an optional token cache.

    Args:
        auth: Authentication credentials - path to credentials file or (client_id, client_secret) tuple
        auth_host: Authentication server base URL
        auth_token_endpoint: Relative path to token endpoint
        token_cache: Token cache for cross-process persistence. When given, a valid cached token is
            injected on startup and new tokens are saved automatically.

    Returns:
        Configured RequestsAuthSession
    """
    client_id, client_secret = resolve_credentials(auth)
    return RequestsAuthSession(
        auth=(client_id, client_secret),
        host=auth_host,
        token_endpoint=auth_token_endpoint,
        initial_token=token_cache.load(auth_host, client_id) if (token_cache and client_id) else None,
        on_token_updated=(lambda t: token_cache.save(auth_host, client_id, t)) if (token_cache and client_id) else None,
    )


_provider_pool: WeakValueDictionary[tuple, RequestsAuthSession] = WeakValueDictionary()
_provider_pool_lock = threading.Lock()


def _get_shared_provider(
    auth: ANY_AUTH_TYPE,
    auth_host: str,
    auth_token_endpoint: str,
    token_cache: Optional[TokenCache] = None,
) -> RequestsAuthSession:
    """Return a shared RequestsAuthSession for the given credentials, creating one if needed.

    Providers are keyed by (client_id, auth_host, auth_token_endpoint, cache_type) and held
    weakly, so they are GC'd once no BaseApiClient instances reference them.
    """
    client_id, client_secret = resolve_credentials(auth)

    if not client_id or not client_secret:
        return RequestsAuthSession(auth=auth, host=auth_host, token_endpoint=auth_token_endpoint)

    key = (client_id, auth_host, auth_token_endpoint, type(token_cache))
    with _provider_pool_lock:
        provider = _provider_pool.get(key)
        if provider is None:
            provider = RequestsAuthSession(
                auth=(client_id, client_secret),
                host=auth_host,
                token_endpoint=auth_token_endpoint,
                initial_token=token_cache.load(auth_host, client_id) if token_cache else None,
                on_token_updated=(lambda t: token_cache.save(auth_host, client_id, t)) if token_cache else None,
            )
            _provider_pool[key] = provider
    return provider


class BaseApiClient:
    """Base API client with OAuth2 authentication using requests.

    Provides a requests Session with:
    - OAuth2 authentication with automatic token refresh
    - Automatic JSON serialization
    - Retry logic for transient errors (502, 503, 504)
    - Sunset header handling
    - Enhanced error messages

    Clients with the same credentials and auth host automatically share a token provider,
    so only one token fetch occurs across all instances with those credentials.

    The interface is consistent with requests - use session.get(), session.post(), etc.
    Calls return the response object. Use response.json() to get the data.

    Example:
        client = BaseApiClient()
        response = client.session.get("https://api.app.kognic.com/v1/resources")
        data = response.json()
    """

    def __init__(
        self,
        *,
        auth: Optional[Union[str, os.PathLike, tuple]] = None,
        auth_host: str = DEFAULT_HOST,
        auth_token_endpoint: str = DEFAULT_TOKEN_ENDPOINT_RELPATH,
        client_name: Optional[str] = "auto",
        json_serializer: Callable[[Any], Any] = serialize_body,
        token_provider: Optional[RequestsAuthSession] = None,
        token_cache: Optional[TokenCache] = None,
        sunset_handler: Optional[SunsetHandler] = _DEFAULT_SUNSET_HANDLER,
    ):
        """Initialize the API client.

        Args:
            auth: Authentication credentials - path to credentials file or (client_id, client_secret) tuple
            auth_host: Authentication server base URL
            auth_token_endpoint: Relative path to token endpoint
            client_name: Name added to User-Agent. Use "auto" for class name, None for no name.
            json_serializer: Callable to serialize request bodies. Defaults to serialize_body.
            token_provider: Explicit token provider to share across clients. When omitted, a shared
                provider is looked up (or created) by credentials + auth_host.
            token_cache: Token cache for cross-process token persistence. When given, a valid cached
                token is injected on startup and new tokens are saved automatically.
            sunset_handler: Callable invoked with ``(sunset_date, method, url)`` when a sunset header
                is detected. Defaults to logging a warning or error. Pass ``None`` to disable.
        """
        self._session: Optional[Session] = None
        self._auth = auth
        self._auth_host = auth_host
        self._auth_token_endpoint = auth_token_endpoint
        self._json_serializer = json_serializer
        self._token_provider = token_provider
        self._token_cache = token_cache
        self._sunset_handler = sunset_handler
        self._lock = Lock()

        if client_name == "auto":
            client_name = self.__class__.__name__
        self._client_name = client_name

    @property
    def session(self) -> Session:
        """Get the authenticated requests Session.

        Session is lazily initialized on first access.
        """
        if self._session is None:
            with self._lock:
                if self._session is None:
                    provider = self._token_provider or _get_shared_provider(
                        self._auth, self._auth_host, self._auth_token_endpoint, self._token_cache
                    )
                    self._session = create_session(
                        token_provider=provider,
                        client_name=self._client_name,
                        json_serializer=self._json_serializer,
                        sunset_handler=self._sunset_handler,
                    )
        return self._session

    @classmethod
    def from_env(
        cls,
        env: str,
        *,
        env_config_path: Union[str, os.PathLike] = "",
        **kwargs,
    ) -> Self:
        """Create a client from a named environment in the config file.

        Args:
            env: Environment name to look up in the config file.
            env_config_path: Path to config file. Defaults to ~/.config/kognic/environments.json.
            **kwargs: Additional arguments passed to the constructor (e.g. client_name, json_serializer).

        Returns:
            Configured BaseApiClient instance.
        """

        config_file_path = env_config_path or DEFAULT_ENV_CONFIG_FILE_PATH
        cfg = load_kognic_env_config(config_file_path)
        if env not in cfg.environments:
            raise ValueError(f"Unknown environment: {env} not found in config at {config_file_path}")
        resolved = cfg.environments[env]
        kwargs.setdefault("auth", resolved.credentials)
        kwargs["auth_host"] = resolved.auth_server
        return cls(**kwargs)



import logging
import os
from logging import NullHandler
from pathlib import Path

from kognic.auth._sunset import SunsetHandler, default_sunset_handler

logging.getLogger(__name__).addHandler(NullHandler())

__all__ = ["SunsetHandler", "default_sunset_handler"]

try:
    from ._version import __version__
except ImportError:
    __version__ = "0.0.0"

DEFAULT_HOST = "https://auth.app.kognic.com"
DEFAULT_KOGNIC_PLATFORM = "app.kognic.com"
DEFAULT_TOKEN_ENDPOINT_RELPATH = "/v1/auth/oauth/token"

DEFAULT_ENV_CONFIG_FILE_PATH = (
    Path(os.environ.get("XDG_CONFIG_HOME", Path.home() / ".config")) / "kognic" / "environments.json"
)

DEFAULT_CACHE_PATH = Path(os.environ.get("XDG_CACHE_HOME", Path.home() / ".cache")) / "kognic-auth" / "tokens.json"



"""Sunset header handling for deprecated API endpoints."""

import logging
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Callable, Optional, Union

if TYPE_CHECKING:
    from ._protocols import Response, Url

SUNSET_HEADER = "sunset-date"

# Expected formats of sunset date
DATETIME_FMT = "%Y-%m-%dT%H:%M:%S.%fZ"
DATETIME_FMT_NO_MICRO = "%Y-%m-%dT%H:%M:%SZ"

logger = logging.getLogger(__name__)

SunsetHandler = Callable[[datetime, str, str], None]


def default_sunset_handler(threshold_days: int = 14) -> SunsetHandler:
    """Return a sunset handler that logs a warning or error based on time until sunset.

    Args:
        threshold_days: Days remaining before the sunset date at which the log level
            escalates from warning to error. Defaults to 14.

    Returns:
        A callable ``(sunset_date, method, url) -> None`` that logs the deprecation notice.

    Example::

        from kognic.auth import default_sunset_handler
        client = BaseApiClient(sunset_handler=default_sunset_handler(threshold_days=30))
    """
    threshold_seconds = threshold_days * 60 * 60 * 24

    def handler(sunset_date: datetime, method: str, url: str) -> None:
        now = datetime.now(tz=timezone.utc)
        diff = sunset_date - now
        log_method = logger.warning if diff.total_seconds() > threshold_seconds else logger.error
        log_method(
            f"Endpoint has been deprecated and will be removed at {sunset_date}. Please update your client. "
            f"Endpoint: {method} {url}"
        )

    return handler


_default_handler: SunsetHandler = default_sunset_handler()


def handle_sunset(response: "Response", handler: Optional[SunsetHandler] = _default_handler) -> None:
    """Check for Sunset header and invoke the handler if present.

    Args:
        response: The HTTP response object (requests.Response or httpx.Response)
        handler: Callable invoked with ``(sunset_date, method, url)``. Pass ``None`` to disable.
    """
    if handler is None:
        return
    sunset_string = response.headers.get(SUNSET_HEADER)
    sunset_date = _parse_date(sunset_string) if sunset_string else None
    if not sunset_date:
        return

    handler(sunset_date, response.request.method, _parse_url(response.request.url))


def _parse_date(date: str) -> Optional[datetime]:
    """Parse sunset date string to UTC datetime."""
    for fmt in (DATETIME_FMT, DATETIME_FMT_NO_MICRO):
        try:
            return datetime.strptime(date, fmt).replace(tzinfo=timezone.utc)
        except ValueError:
            continue
    return None


def _parse_url(url: Union[str, "Url"]) -> str:
    """Extract clean URL without query parameters."""
    if isinstance(url, str):
        return url.split("?")[0]
    return f"{url.scheme}://{url.host}{url.path}"



"""Tests for the keyring-based credential store."""

import json
import unittest
from unittest import mock

from kognic.auth.credentials_parser import ApiCredentials
from kognic.auth.internal.credentials_store import (
    DEFAULT_PROFILE,
    SERVICE_NAME,
    clear_credentials,
    load_credentials,
    save_credentials,
)

FULL_CREDS_DICT = {
    "clientId": "my-id",
    "clientSecret": "my-secret",
    "email": "test@kognic.com",
    "userId": 1,
    "issuer": "auth.kognic.test",
}

FULL_CREDS = ApiCredentials(
    client_id="my-id",
    client_secret="my-secret",
    email="test@kognic.com",
    user_id=1,
    issuer="auth.kognic.test",
)


def _mock_keyring(get_password=None, set_password=None, delete_password=None):
    """Return a mock keyring module wired to the given side effects / return values."""
    kr = mock.MagicMock()
    if get_password is not None:
        kr.get_password.return_value = get_password
    if set_password is not None:
        kr.set_password.side_effect = set_password
    if delete_password is not None:
        kr.delete_password.side_effect = delete_password
    kr.get_keyring.return_value = mock.MagicMock()  # non-fail backend
    return kr


class LoadCredentialsTest(unittest.TestCase):
    def test_no_keyring_returns_none(self):
        with mock.patch("kognic.auth.internal.credentials_store._get_keyring", return_value=None):
            self.assertIsNone(load_credentials())

    def test_not_stored_returns_none(self):
        kr = _mock_keyring(get_password=None)
        with mock.patch("kognic.auth.internal.credentials_store._get_keyring", return_value=kr):
            self.assertIsNone(load_credentials())
        kr.get_password.assert_called_once_with(SERVICE_NAME, DEFAULT_PROFILE)

    def test_stored_credentials_returned(self):
        data = json.dumps(FULL_CREDS_DICT)
        kr = _mock_keyring(get_password=data)
        with mock.patch("kognic.auth.internal.credentials_store._get_keyring", return_value=kr):
            result = load_credentials()
        self.assertIsInstance(result, ApiCredentials)
        self.assertEqual(result.client_id, "my-id")
        self.assertEqual(result.client_secret, "my-secret")
        self.assertEqual(result.email, "test@kognic.com")
        self.assertEqual(result.user_id, 1)
        self.assertEqual(result.issuer, "auth.kognic.test")

    def test_custom_profile(self):
        data = json.dumps(FULL_CREDS_DICT)
        kr = _mock_keyring(get_password=data)
        with mock.patch("kognic.auth.internal.credentials_store._get_keyring", return_value=kr):
            load_credentials(profile="demo")
        kr.get_password.assert_called_once_with(SERVICE_NAME, "demo")

    def test_corrupt_json_returns_none(self):
        kr = _mock_keyring(get_password="not-json")
        with mock.patch("kognic.auth.internal.credentials_store._get_keyring", return_value=kr):
            self.assertIsNone(load_credentials())

    def test_keyring_error_returns_none(self):
        kr = mock.MagicMock()
        kr.get_password.side_effect = Exception("keyring exploded")
        with mock.patch("kognic.auth.internal.credentials_store._get_keyring", return_value=kr):
            self.assertIsNone(load_credentials())


class SaveCredentialsTest(unittest.TestCase):
    def test_no_keyring_raises(self):
        with mock.patch("kognic.auth.internal.credentials_store._get_keyring", return_value=None):
            with self.assertRaises(RuntimeError) as ctx:
                save_credentials(FULL_CREDS)
            self.assertIn("keyring", str(ctx.exception).lower())

    def test_stores_in_keyring(self):
        kr = mock.MagicMock()
        with mock.patch("kognic.auth.internal.credentials_store._get_keyring", return_value=kr):
            save_credentials(FULL_CREDS)
        kr.set_password.assert_called_once_with(
            SERVICE_NAME,
            DEFAULT_PROFILE,
            json.dumps(FULL_CREDS_DICT),
        )

    def test_custom_profile(self):
        kr = mock.MagicMock()
        with mock.patch("kognic.auth.internal.credentials_store._get_keyring", return_value=kr):
            save_credentials(FULL_CREDS, profile="demo")
        kr.set_password.assert_called_once_with(
            SERVICE_NAME,
            "demo",
            json.dumps(FULL_CREDS_DICT),
        )


class ClearCredentialsTest(unittest.TestCase):
    def test_no_keyring_does_not_raise(self):
        with mock.patch("kognic.auth.internal.credentials_store._get_keyring", return_value=None):
            clear_credentials()  # should not raise

    def test_clears_from_keyring(self):
        kr = mock.MagicMock()
        with mock.patch("kognic.auth.internal.credentials_store._get_keyring", return_value=kr):
            clear_credentials()
        kr.delete_password.assert_called_once_with(SERVICE_NAME, DEFAULT_PROFILE)

    def test_error_silenced(self):
        kr = mock.MagicMock()
        kr.delete_password.side_effect = Exception("gone already")
        with mock.patch("kognic.auth.internal.credentials_store._get_keyring", return_value=kr):
            clear_credentials()  # should not raise


if __name__ == "__main__":
    unittest.main()



"""Serialization and deserialization utilities for HTTP request/response bodies."""

from typing import Any, Dict, List, Optional, Union

ENVELOPED_KEY = "data"


def serialize_body(body: Any) -> Any:
    """Serialize request body to JSON-compatible format.

    Supports:
    - None, dict, list, primitives (passed through)

    Raises:
        ValueError: If body is str or bytes at top level (not supported as request body)
        TypeError: If body type is not supported
    """
    if body is None:
        return None
    if isinstance(body, (str, bytes)):
        raise ValueError("str and bytes data is not supported")
    return _serialize_value(body)


def _serialize_value(value: Any) -> Any:
    """Recursively serialize a value (used internally for container contents)."""
    if value is None:
        return None
    if isinstance(value, (str, int, float, bool)):
        return value
    if isinstance(value, bytes):
        raise ValueError("bytes data is not supported")
    if isinstance(value, dict):
        return {k: _serialize_value(v) for k, v in value.items()}
    if isinstance(value, list):
        return [_serialize_value(item) for item in value]
    raise TypeError(f"Cannot serialize value of type {type(value).__name__}. Expected dict, list, or primitive.")


def deserialize(
    response: Union[Any, Dict[str, Any], List],
    enveloped_key: Optional[str] = ENVELOPED_KEY,
) -> Any:
    """Deserialize a response from the API.

    Designed to work with httpx and requests response objects by duck typing.

    Args:
        response: Response object (with .json() method) or dict/list
        enveloped_key: By Kognic convention, data is enveloped in a key.
            Default is 'data'. Set to None to skip envelope extraction.

    Returns:
        Deserialized data as raw dict/list

    Raises:
        ValueError: If enveloped_key is specified but not found in response
    """
    # Extract JSON from response object or use directly if already a dict/list
    try:
        response_json = response.json()
    except AttributeError:
        response_json = response

    # Extract data from envelope if specified
    if enveloped_key is not None:
        if enveloped_key not in response_json:
            raise ValueError(
                f"Expected enveloped key '{enveloped_key}' not found in response json. "
                f"Found keys: {response_json.keys()}"
            )
        return response_json[enveloped_key]

    return response_json



import json
import tempfile
import unittest
from pathlib import Path

from kognic.auth import DEFAULT_HOST
from kognic.auth.env_config import Environment, KognicEnvConfig, load_kognic_env_config, resolve_environment


class LoadConfigTest(unittest.TestCase):
    def test_missing_file_returns_empty_config(self):
        config = load_kognic_env_config("/nonexistent/path/config.json")
        self.assertEqual(config.environments, {})
        self.assertIsNone(config.default_environment)

    def test_valid_config(self):
        data = {
            "environments": {
                "production": {
                    "host": "app.kognic.com",
                    "auth_server": "https://auth.app.kognic.com",
                    "credentials": "~/creds.json",
                },
                "demo": {
                    "host": "demo.kognic.com",
                    "auth_server": "https://auth.demo.kognic.com",
                },
            },
            "default_environment": "production",
        }
        with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as f:
            json.dump(data, f)
            f.flush()
            config = load_kognic_env_config(f.name)

        Path(f.name).unlink()

        self.assertEqual(len(config.environments), 2)
        self.assertEqual(config.default_environment, "production")

        prod = config.environments["production"]
        self.assertEqual(prod.name, "production")
        self.assertEqual(prod.host, "app.kognic.com")
        self.assertEqual(prod.auth_server, "https://auth.app.kognic.com")
        self.assertTrue(prod.credentials.endswith("creds.json"))
        self.assertNotIn("~", prod.credentials)

        demo = config.environments["demo"]
        self.assertIsNone(demo.credentials)

    def test_invalid_json_raises(self):
        with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as f:
            f.write("not valid json{")
            f.flush()
            with self.assertRaises(json.JSONDecodeError):
                load_kognic_env_config(f.name)
        Path(f.name).unlink()

    def test_keyring_uri_not_expanded(self):
        """keyring:// credentials are stored as-is, not treated as file paths."""
        data = {
            "environments": {
                "production": {
                    "host": "app.kognic.com",
                    "auth_server": "https://auth.app.kognic.com",
                    "credentials": "keyring://production",
                }
            }
        }
        with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as f:
            json.dump(data, f)
            f.flush()
            config = load_kognic_env_config(f.name)
        Path(f.name).unlink()

        self.assertEqual(config.environments["production"].credentials, "keyring://production")

    def test_empty_contexts(self):
        data = {"environments": {}}
        with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as f:
            json.dump(data, f)
            f.flush()
            config = load_kognic_env_config(f.name)
        Path(f.name).unlink()

        self.assertEqual(config.environments, {})
        self.assertIsNone(config.default_environment)


class ResolveEnvironmentTest(unittest.TestCase):
    def setUp(self):
        self.config = KognicEnvConfig(
            environments={
                "production": Environment(
                    name="production",
                    host="app.kognic.com",
                    auth_server="https://auth.app.kognic.com",
                    credentials="/path/to/prod-creds.json",
                ),
                "demo": Environment(
                    name="demo",
                    host="demo.kognic.com",
                    auth_server="https://auth.demo.kognic.com",
                    credentials="/path/to/demo-creds.json",
                ),
            },
            default_environment="production",
        )

    def test_explicit_env(self):
        env = resolve_environment(self.config, "https://anything.com/v1/foo", "demo")
        self.assertEqual(env.name, "demo")

    def test_explicit_env_unknown_raises(self):
        with self.assertRaises(ValueError) as cm:
            resolve_environment(self.config, "https://anything.com", "nonexistent")
        self.assertIn("Unknown environment", str(cm.exception))

    def test_exact_host_match(self):
        env = resolve_environment(self.config, "https://app.kognic.com/v1/projects")
        self.assertEqual(env.name, "production")

    def test_subdomain_match(self):
        env = resolve_environment(self.config, "https://api.app.kognic.com/v1/projects")
        self.assertEqual(env.name, "production")

    def test_demo_exact_match(self):
        env = resolve_environment(self.config, "https://demo.kognic.com/v1/projects")
        self.assertEqual(env.name, "demo")

    def test_demo_subdomain_match(self):
        env = resolve_environment(self.config, "https://api.demo.kognic.com/v1/projects")
        self.assertEqual(env.name, "demo")

    def test_default_environment_fallback(self):
        env = resolve_environment(self.config, "https://unknown.example.com/v1/foo")
        self.assertEqual(env.name, "production")

    def test_no_config_fallback(self):
        empty_config = KognicEnvConfig()
        env = resolve_environment(empty_config, "https://app.kognic.com/v1/projects")
        self.assertEqual(env.name, "default")
        self.assertEqual(env.auth_server, DEFAULT_HOST)
        self.assertIsNone(env.credentials)

    def test_no_default_no_match_falls_back_to_default_auth(self):
        config = KognicEnvConfig(
            environments={
                "demo": Environment(
                    name="demo",
                    host="demo.kognic.com",
                    auth_server="https://auth.demo.kognic.com",
                ),
            },
            default_environment=None,
        )
        env = resolve_environment(config, "https://unknown.example.com/v1/foo")
        self.assertEqual(env.name, "default")
        self.assertEqual(env.auth_server, DEFAULT_HOST)


if __name__ == "__main__":
    unittest.main()



"""Tests for requests-layer token refresh and 401 retry behaviour."""

import unittest
from unittest.mock import MagicMock, patch

import requests
from authlib.common.errors import AuthlibBaseError

from kognic.auth.requests.auth_session import _FixedSession
from kognic.auth.requests.bearer_auth import KognicBearerAuth


def _make_fixed_session():
    """Return a _FixedSession with dummy credentials and no network calls."""
    return _FixedSession(
        client_id="test-id",
        client_secret="test-secret",
        token_endpoint_auth_method="client_secret_post",
        grant_type="client_credentials",
    )


class TestKognicBearerAuthCall(unittest.TestCase):
    def test_injects_bearer_token(self):
        provider = MagicMock()
        provider.ensure_token.return_value = {"access_token": "tok-abc"}

        auth = KognicBearerAuth(provider)
        prepared = MagicMock()
        prepared.headers = {}

        result = auth(prepared)

        self.assertEqual(prepared.headers["Authorization"], "Bearer tok-abc")
        prepared.register_hook.assert_called_once_with("response", auth._handle_401)
        self.assertIs(result, prepared)


class TestKognicBearerAuthHandle401(unittest.TestCase):
    def _make_auth(self, new_token="fresh-token"):
        provider = MagicMock()
        provider.ensure_token.return_value = {"access_token": new_token}
        return KognicBearerAuth(provider), provider

    def _make_401_response(self):
        # Don't use spec=requests.Response; `connection` is an internal attribute
        # not listed in the public API so spec would block it.
        resp = MagicMock()
        resp.status_code = 401
        resp.content = b""
        copied_prep = MagicMock()
        copied_prep.headers = {}  # real dict so __setitem__ is observable
        resp.request.copy.return_value = copied_prep
        retry_resp = MagicMock()
        retry_resp.history = []
        resp.connection.send.return_value = retry_resp
        return resp, retry_resp

    def test_non_401_passes_through(self):
        auth, provider = self._make_auth()
        resp = MagicMock(spec=requests.Response)
        resp.status_code = 200

        result = auth._handle_401(resp)

        self.assertIs(result, resp)
        provider.invalidate_token.assert_not_called()

    def test_401_invalidates_token(self):
        auth, provider = self._make_auth()
        resp, _ = self._make_401_response()

        auth._handle_401(resp)

        provider.invalidate_token.assert_called_once()

    def test_401_retries_with_new_token(self):
        auth, provider = self._make_auth("brand-new-token")
        resp, retry_resp = self._make_401_response()

        result = auth._handle_401(resp)

        # ensure_token called twice: once in __call__ (not here), once in _handle_401
        provider.ensure_token.assert_called_once()
        copied = resp.request.copy.return_value
        self.assertEqual(copied.headers["Authorization"], "Bearer brand-new-token")
        resp.connection.send.assert_called_once_with(copied)
        self.assertIs(result, retry_resp)

    def test_401_original_response_appended_to_history(self):
        auth, _ = self._make_auth()
        resp, retry_resp = self._make_401_response()

        auth._handle_401(resp)

        self.assertIn(resp, retry_resp.history)

    def test_401_passes_kwargs_to_send(self):
        auth, _ = self._make_auth()
        resp, retry_resp = self._make_401_response()

        auth._handle_401(resp, timeout=5, verify=False)

        _, kwargs = resp.connection.send.call_args
        self.assertEqual(kwargs.get("timeout"), 5)
        self.assertFalse(kwargs.get("verify"))


class TestFixedSessionRefreshToken(unittest.TestCase):
    def test_authlib_invalid_token_calls_fetch_token(self):
        session = _make_fixed_session()
        err = AuthlibBaseError(error="invalid_token", description="expired")

        with patch.object(session.__class__.__bases__[0], "refresh_token", side_effect=err):
            with patch.object(session, "fetch_token", return_value={"access_token": "new"}) as mock_fetch:
                session.refresh_token("https://auth.kognic.com/token")

        mock_fetch.assert_called_once()

    def test_authlib_other_error_reraises(self):
        session = _make_fixed_session()
        err = AuthlibBaseError(error="server_error", description="oops")

        with patch.object(session.__class__.__bases__[0], "refresh_token", side_effect=err):
            with self.assertRaises(AuthlibBaseError) as cm:
                session.refresh_token("https://auth.kognic.com/token")

        self.assertEqual(cm.exception.error, "server_error")

    def test_http_401_invalid_token_calls_fetch_token(self):
        session = _make_fixed_session()
        http_resp = MagicMock(spec=requests.Response)
        http_resp.status_code = 401
        http_resp.json.return_value = {"error": "invalid_token"}
        http_err = requests.exceptions.HTTPError(response=http_resp)

        with patch.object(session.__class__.__bases__[0], "refresh_token", side_effect=http_err):
            with patch.object(session, "fetch_token", return_value={"access_token": "new"}) as mock_fetch:
                session.refresh_token("https://auth.kognic.com/token")

        mock_fetch.assert_called_once()

    def test_http_401_other_error_reraises(self):
        session = _make_fixed_session()
        http_resp = MagicMock(spec=requests.Response)
        http_resp.status_code = 401
        http_resp.json.return_value = {"error": "access_denied"}
        http_err = requests.exceptions.HTTPError(response=http_resp)

        with patch.object(session.__class__.__bases__[0], "refresh_token", side_effect=http_err):
            with self.assertRaises(requests.exceptions.HTTPError):
                session.refresh_token("https://auth.kognic.com/token")

    def test_http_non_401_reraises(self):
        session = _make_fixed_session()
        http_resp = MagicMock(spec=requests.Response)
        http_resp.status_code = 500
        http_resp.json.return_value = {"error": "invalid_token"}
        http_err = requests.exceptions.HTTPError(response=http_resp)

        with patch.object(session.__class__.__bases__[0], "refresh_token", side_effect=http_err):
            with self.assertRaises(requests.exceptions.HTTPError):
                session.refresh_token("https://auth.kognic.com/token")


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import json
import logging
from typing import Optional

from kognic.auth.internal.token_cache._base import SERVICE_NAME, TokenCache, is_valid, make_key

log = logging.getLogger(__name__)

_KEYRING_MISSING = object()  # sentinel: import attempted but unavailable


class KeyringTokenCache(TokenCache):
    """Token cache backed by the system keyring."""

    def __init__(self) -> None:
        self._keyring_module = None  # not yet resolved

    def _keyring(self):
        """Return the keyring module if usable, else None. Result is cached."""
        if self._keyring_module is _KEYRING_MISSING:
            return None
        if self._keyring_module is not None:
            return self._keyring_module
        try:
            import keyring

            backend = keyring.get_keyring()
            if "fail" in type(backend).__name__.lower():
                raise RuntimeError("unusable keyring backend")
            self._keyring_module = keyring
        except Exception:
            self._keyring_module = _KEYRING_MISSING
            return None
        return self._keyring_module

    def load(self, auth_server: str, client_id: str) -> Optional[dict]:
        kr = self._keyring()
        if kr is None:
            return None
        try:
            key = make_key(auth_server, client_id)
            stored = kr.get_password(SERVICE_NAME, key)
            if stored is None:
                return None
            token = json.loads(stored)
            if not is_valid(token):
                log.debug("Cached keyring token expired or missing expires_at, discarding")
                return None
            log.debug("Using cached token from keyring (key=%s)", key)
            return token
        except Exception:
            log.debug("Failed to load token from keyring", exc_info=True)
            return None

    def save(self, auth_server: str, client_id: str, token: dict) -> None:
        kr = self._keyring()
        if kr is None:
            return
        try:
            key = make_key(auth_server, client_id)
            kr.set_password(SERVICE_NAME, key, json.dumps(token))
            log.debug("Saved token to keyring for key=%s", key)
        except Exception:
            log.debug("Failed to save token to keyring", exc_info=True)

    def clear(self, auth_server: str, client_id: str) -> None:
        kr = self._keyring()
        if kr is None:
            return
        try:
            key = make_key(auth_server, client_id)
            kr.delete_password(SERVICE_NAME, key)
            log.debug("Cleared cached token from keyring for key=%s", key)
        except Exception:
            log.debug("Failed to clear token from keyring", exc_info=True)



"""Unit tests for serialization utilities."""

import unittest

from kognic.auth.serde import serialize_body


class TestSerializeBody(unittest.TestCase):
    def test_serialize_none(self):
        self.assertIsNone(serialize_body(None))

    def test_serialize_dict(self):
        data = {"key": "value"}
        self.assertEqual(serialize_body(data), data)

    def test_serialize_nested_dict(self):
        data = {"key": {"nested_key": "value"}}
        self.assertEqual(serialize_body(data), data)

    def test_serialize_list(self):
        data = ["apa", "bepa"]
        self.assertEqual(serialize_body(data), data)

    def test_serialize_str_raises(self):
        with self.assertRaises(ValueError):
            serialize_body('{"key": "value"}')

    def test_serialize_bytes_raises(self):
        with self.assertRaises(ValueError):
            serialize_body(b'{"key": "value"}')

    def test_serialize_unsupported_type_raises(self):
        class UnsupportedType:
            pass

        with self.assertRaises(TypeError):
            serialize_body(UnsupportedType())

    def test_serialize_dict_with_strings(self):
        data = {"key": "value", "nested": {"inner": "string"}}
        self.assertEqual(serialize_body(data), data)

    def test_serialize_list_with_strings(self):
        data = ["a", "b", "c"]
        self.assertEqual(serialize_body(data), data)

    def test_serialize_mixed_list(self):
        data = ["string", 42, {"key": "value"}]
        result = serialize_body(data)
        self.assertEqual(result, ["string", 42, {"key": "value"}])


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import argparse
import sys

from kognic.auth import DEFAULT_HOST
from kognic.auth.env_config import DEFAULT_ENV_CONFIG_FILE_PATH, load_kognic_env_config
from kognic.auth.internal.token_cache import make_cache
from kognic.auth.requests.base_client import make_token_provider

COMMAND = "get-access-token"


def register_parser(subparsers: argparse._SubParsersAction) -> argparse.ArgumentParser:
    token_parser = subparsers.add_parser(
        COMMAND,
        help="Generate an access token for Kognic API authentication",
    )
    token_parser.add_argument(
        "--server",
        default=None,
        help=f"Authentication server URL (default: {DEFAULT_HOST})",
    )
    token_parser.add_argument(
        "--credentials",
        metavar="FILE",
        help="Path to JSON credentials file. If not provided, credentials are read from environment variables.",
    )
    token_parser.add_argument(
        "--env-config-file-path",
        default=DEFAULT_ENV_CONFIG_FILE_PATH,
        help=f"Environment config file path (default: {DEFAULT_ENV_CONFIG_FILE_PATH})",
    )
    token_parser.add_argument(
        "--env",
        dest="env_name",
        help="Use a specific environment from the config file",
    )
    token_parser.add_argument(
        "--token-cache",
        choices=["auto", "keyring", "file", "none"],
        default="auto",
        help="Token cache backend: auto (default), keyring, file, or none. "
        "Auto will use keyring if available, otherwise file-based caching.",
    )

    return token_parser


def run(parsed: argparse.Namespace) -> int:
    try:
        host = parsed.server
        credentials = parsed.credentials

        if parsed.env_name:
            config = load_kognic_env_config(parsed.env_config_file_path)
            if parsed.env_name not in config.environments:
                raise ValueError(
                    f"Environment '{parsed.env_name}' not found in config file '{parsed.env_config_file_path}'"
                )
            ctx = config.environments[parsed.env_name]
            if host is None:
                host = ctx.auth_server
            if credentials is None:
                credentials = ctx.credentials

        auth_host = host or DEFAULT_HOST

        provider = make_token_provider(
            auth=credentials,
            auth_host=auth_host,
            token_cache=make_cache(parsed.token_cache),
        )
        print(provider.ensure_token()["access_token"])
        return 0
    except FileNotFoundError as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1
    except ValueError as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1
    except Exception as e:
        print(f"Error fetching token: {e}", file=sys.stderr)
        return 1



"""Token cache backends for cross-process token persistence.

All keyring imports are lazy so the module works when keyring is not installed.
"""

from __future__ import annotations

from kognic.auth.internal.token_cache._base import TokenCache
from kognic.auth.internal.token_cache._file import FileTokenCache
from kognic.auth.internal.token_cache._keyring import KeyringTokenCache


def make_cache(mode: str) -> TokenCache | None:
    """Return a TokenCache for the given mode, or None for 'none'.

    Modes:
      auto    – keyring if available, file otherwise (default)
      keyring – system keyring only
      file    – file-based cache only
      none    – no caching
    """
    if mode == "none":
        return None
    if mode == "file":
        return FileTokenCache()
    if mode == "keyring":
        return KeyringTokenCache()
    # auto
    candidate = KeyringTokenCache()
    if candidate._keyring() is not None:
        return candidate
    return FileTokenCache()


__all__ = ["TokenCache", "KeyringTokenCache", "FileTokenCache", "make_cache"]



from __future__ import annotations

import csv
import io
import json
from typing import Any


def _extract_items(body: Any) -> list[Any] | None:
    if isinstance(body, list):
        return body
    if isinstance(body, dict):
        values = list(body.values())
        if len(values) == 1 and isinstance(values[0], list):
            return values[0]
    return None


def _stringify_value(value: Any) -> str:
    if isinstance(value, (dict, list)):
        return json.dumps(value)
    return str(value)


def _collect_fieldnames(items: list[Any]) -> list[str]:
    fieldnames: list[str] = []
    for item in items:
        if isinstance(item, dict):
            for key in item:
                if key not in fieldnames:
                    fieldnames.append(key)
    return fieldnames


def _print_delimited(items: list[Any], *, delimiter: str = ",") -> None:
    if not items:
        return
    fieldnames = _collect_fieldnames(items)
    buf = io.StringIO()
    writer = csv.DictWriter(buf, fieldnames=fieldnames, delimiter=delimiter)
    writer.writeheader()
    for item in items:
        if isinstance(item, dict):
            writer.writerow({k: _stringify_value(v) for k, v in item.items()})
        else:
            writer.writerow({"value": _stringify_value(item)})
    print(buf.getvalue(), end="")


def _print_table(items: list[Any]) -> None:
    if not items:
        return
    fieldnames = _collect_fieldnames(items)
    col_widths = [len(f) for f in fieldnames]
    rows: list[list[str]] = []
    for item in items:
        row = [
            _stringify_value(item.get(f, "")) if isinstance(item, dict) else _stringify_value(item) for f in fieldnames
        ]
        rows.append(row)
        for i, cell in enumerate(row):
            col_widths[i] = max(col_widths[i], len(cell))
    header = "| " + " | ".join(f.ljust(col_widths[i]) for i, f in enumerate(fieldnames)) + " |"
    separator = "|-" + "-|-".join("-" * w for w in col_widths) + "-|"
    print(header)
    print(separator)
    for row in rows:
        print("| " + " | ".join(cell.ljust(col_widths[i]) for i, cell in enumerate(row)) + " |")


def _print_response(response: Any, *, output_format: str = "json") -> None:
    content_type = response.headers.get("Content-Type", "")
    if "application/json" in content_type:
        try:
            body = response.json()
        except (json.JSONDecodeError, ValueError):
            print(response.text)
            return
        if output_format in ("jsonl", "csv", "tsv", "table"):
            items = _extract_items(body)
            if items is not None:
                if output_format in ("csv", "tsv"):
                    _print_delimited(items, delimiter="\t" if output_format == "tsv" else ",")
                elif output_format == "table":
                    _print_table(items)
                else:
                    for item in items:
                        print(json.dumps(item))
                return
        print(json.dumps(body, indent=2))
    else:
        print(response.text)



import json
import time
import unittest
from unittest import mock

from kognic.auth.internal.token_cache import KeyringTokenCache
from kognic.auth.internal.token_cache._base import EXPIRY_MARGIN_SECONDS, SERVICE_NAME, make_key


def _make_token(*, expires_in=3600, extra=None):
    """Create a realistic token dict."""
    now = time.time()
    token = {
        "access_token": "eyJ.test.token",
        "token_type": "bearer",
        "expires_in": expires_in,
        "expires_at": now + expires_in,
    }
    if extra:
        token.update(extra)
    return token


def _cache_no_keyring() -> KeyringTokenCache:
    cache = KeyringTokenCache()
    cache._keyring = lambda: None
    return cache


def _cache_with_keyring(mock_kr) -> KeyringTokenCache:
    cache = KeyringTokenCache()
    cache._keyring = lambda: mock_kr
    return cache


class MakeKeyTest(unittest.TestCase):
    def test_format(self):
        key = make_key("https://auth.app.kognic.com", "my-client-id")
        self.assertEqual(key, "https://auth.app.kognic.com:my-client-id")

    def test_different_servers_produce_different_keys(self):
        key1 = make_key("https://auth.app.kognic.com", "client-1")
        key2 = make_key("https://auth.demo.kognic.com", "client-1")
        self.assertNotEqual(key1, key2)


class KeyringAvailableTest(unittest.TestCase):
    def test_keyring_available_when_valid_backend(self):
        cache = KeyringTokenCache()
        mock_kr = mock.MagicMock()
        mock_kr.get_keyring.return_value = mock.MagicMock()
        with mock.patch.dict("sys.modules", {"keyring": mock_kr}):
            result = cache._keyring()
        self.assertIsNotNone(result)

    def test_keyring_unavailable_when_fail_backend(self):
        cache = KeyringTokenCache()

        class FailKeyring:
            pass

        mock_kr = mock.MagicMock()
        mock_kr.get_keyring.return_value = FailKeyring()
        with mock.patch.dict("sys.modules", {"keyring": mock_kr}):
            result = cache._keyring()
        self.assertIsNone(result)


class LoadCachedTokenTest(unittest.TestCase):
    def test_keyring_not_available(self):
        result = _cache_no_keyring().load("https://auth.app.kognic.com", "client-1")
        self.assertIsNone(result)

    def test_not_found(self):
        mock_kr = mock.MagicMock()
        mock_kr.get_password.return_value = None
        result = _cache_with_keyring(mock_kr).load("https://auth.app.kognic.com", "client-1")
        self.assertIsNone(result)
        mock_kr.get_password.assert_called_once_with(SERVICE_NAME, "https://auth.app.kognic.com:client-1")

    def test_valid_token(self):
        token = _make_token(expires_in=3600)
        mock_kr = mock.MagicMock()
        mock_kr.get_password.return_value = json.dumps(token)
        result = _cache_with_keyring(mock_kr).load("https://auth.app.kognic.com", "client-1")
        self.assertIsNotNone(result)
        self.assertEqual(result["access_token"], "eyJ.test.token")

    def test_expired_token(self):
        token = _make_token(expires_in=-100)
        mock_kr = mock.MagicMock()
        mock_kr.get_password.return_value = json.dumps(token)
        result = _cache_with_keyring(mock_kr).load("https://auth.app.kognic.com", "client-1")
        self.assertIsNone(result)

    def test_token_within_margin(self):
        token = _make_token(expires_in=EXPIRY_MARGIN_SECONDS - 1)
        mock_kr = mock.MagicMock()
        mock_kr.get_password.return_value = json.dumps(token)
        result = _cache_with_keyring(mock_kr).load("https://auth.app.kognic.com", "client-1")
        self.assertIsNone(result)

    def test_no_expires_at(self):
        token = {"access_token": "eyJ.test.token", "token_type": "bearer"}
        mock_kr = mock.MagicMock()
        mock_kr.get_password.return_value = json.dumps(token)
        result = _cache_with_keyring(mock_kr).load("https://auth.app.kognic.com", "client-1")
        self.assertIsNone(result)

    def test_corrupt_json(self):
        mock_kr = mock.MagicMock()
        mock_kr.get_password.return_value = "not valid json!!!"
        result = _cache_with_keyring(mock_kr).load("https://auth.app.kognic.com", "client-1")
        self.assertIsNone(result)

    def test_keyring_error(self):
        mock_kr = mock.MagicMock()
        mock_kr.get_password.side_effect = Exception("keyring error")
        result = _cache_with_keyring(mock_kr).load("https://auth.app.kognic.com", "client-1")
        self.assertIsNone(result)

    def test_includes_refresh_token(self):
        token = _make_token(extra={"refresh_token": "refresh-abc"})
        mock_kr = mock.MagicMock()
        mock_kr.get_password.return_value = json.dumps(token)
        result = _cache_with_keyring(mock_kr).load("https://auth.app.kognic.com", "client-1")
        self.assertIsNotNone(result)
        self.assertEqual(result["refresh_token"], "refresh-abc")


class SaveTokenTest(unittest.TestCase):
    def test_keyring_not_available(self):
        # Should not raise
        _cache_no_keyring().save("https://auth.app.kognic.com", "client-1", _make_token())

    def test_saves_to_keyring(self):
        token = _make_token()
        mock_kr = mock.MagicMock()
        _cache_with_keyring(mock_kr).save("https://auth.app.kognic.com", "client-1", token)
        mock_kr.set_password.assert_called_once_with(
            SERVICE_NAME,
            "https://auth.app.kognic.com:client-1",
            json.dumps(token),
        )

    def test_save_error_silenced(self):
        mock_kr = mock.MagicMock()
        mock_kr.set_password.side_effect = Exception("write failed")
        # Should not raise
        _cache_with_keyring(mock_kr).save("https://auth.app.kognic.com", "client-1", _make_token())


class ClearTokenTest(unittest.TestCase):
    def test_keyring_not_available(self):
        # Should not raise
        _cache_no_keyring().clear("https://auth.app.kognic.com", "client-1")

    def test_clears_from_keyring(self):
        mock_kr = mock.MagicMock()
        _cache_with_keyring(mock_kr).clear("https://auth.app.kognic.com", "client-1")
        mock_kr.delete_password.assert_called_once_with(
            SERVICE_NAME,
            "https://auth.app.kognic.com:client-1",
        )

    def test_clear_error_silenced(self):
        mock_kr = mock.MagicMock()
        mock_kr.delete_password.side_effect = Exception("delete failed")
        # Should not raise
        _cache_with_keyring(mock_kr).clear("https://auth.app.kognic.com", "client-1")


if __name__ == "__main__":
    unittest.main()



"""Unit tests for deserialization utilities."""

import unittest

from httpx import Response

from kognic.auth.serde import deserialize


class TestDeserialize(unittest.TestCase):
    def test_deserialize_to_raw(self):
        resp = Response(200, json={"data": {"key": "value"}})
        val = deserialize(resp)
        self.assertEqual(val, {"key": "value"})

    def test_deserialize_with_custom_envelope_key(self):
        resp = {"custom_key": {"key": "value"}}
        val = deserialize(resp, enveloped_key="custom_key")
        self.assertEqual(val, {"key": "value"})

    def test_deserialize_from_dict(self):
        resp = {"data": {"key": "value"}}
        val = deserialize(resp)
        self.assertEqual(val, {"key": "value"})

    def test_deserialize_list(self):
        resp = Response(200, json={"data": [1, 2, 3]})
        val = deserialize(resp)
        self.assertEqual(val, [1, 2, 3])

    def test_deserialize_list_of_dicts(self):
        resp = Response(200, json={"data": [{"key": "value"}]})
        val = deserialize(resp)
        self.assertEqual(val, [{"key": "value"}])

    def test_deserialize_empty_list(self):
        resp = Response(200, json={"data": []})
        val = deserialize(resp)
        self.assertEqual(val, [])

    def test_deserialize_no_envelope(self):
        resp = Response(200, json={"key": "value"})
        val = deserialize(resp, enveloped_key=None)
        self.assertEqual(val, {"key": "value"})

    def test_deserialize_missing_envelope_key_raises(self):
        resp = Response(200, json={"wrong_key": {"key": "value"}})
        with self.assertRaises(ValueError) as context:
            deserialize(resp)
        self.assertIn("Expected enveloped key 'data' not found", str(context.exception))


if __name__ == "__main__":
    unittest.main()



import json
import time
import unittest
from pathlib import Path
from unittest import mock

from kognic.auth import DEFAULT_HOST
from kognic.auth.cli import create_parser, main
from kognic.auth.cli.api_request import _create_parser as create_kog_parser
from kognic.auth.cli.api_request import run as call_run
from kognic.auth.env_config import Environment


class CliParserTest(unittest.TestCase):
    def test_default_server(self):
        parser = create_parser()
        args = parser.parse_args(["get-access-token"])
        self.assertIsNone(args.server)
        self.assertIsNone(args.credentials)
        self.assertIsNone(args.env_name)

    def test_custom_server(self):
        parser = create_parser()
        args = parser.parse_args(["get-access-token", "--server", "https://custom.auth.server"])
        self.assertEqual(args.server, "https://custom.auth.server")

    def test_credentials_file(self):
        parser = create_parser()
        args = parser.parse_args(["get-access-token", "--credentials", "/path/to/creds.json"])
        self.assertEqual(args.credentials, "/path/to/creds.json")

    def test_all_options(self):
        parser = create_parser()
        args = parser.parse_args(
            ["get-access-token", "--server", "https://my.server", "--credentials", "my_creds.json"]
        )
        self.assertEqual(args.server, "https://my.server")
        self.assertEqual(args.credentials, "my_creds.json")

    def test_get_access_token_with_context(self):
        parser = create_parser()
        args = parser.parse_args(["get-access-token", "--env", "demo"])
        self.assertEqual(args.env_name, "demo")

    def test_no_command_shows_help(self):
        with mock.patch("builtins.print"):
            result = main([])
        self.assertEqual(result, 0)

    def test_token_cache_default(self):
        parser = create_parser()
        args = parser.parse_args(["get-access-token"])
        self.assertEqual(args.token_cache, "auto")

    def test_token_cache_none(self):
        parser = create_parser()
        args = parser.parse_args(["get-access-token", "--token-cache", "none"])
        self.assertEqual(args.token_cache, "none")

    def test_token_cache_choices(self):
        parser = create_parser()
        for choice in ("auto", "keyring", "file", "none"):
            args = parser.parse_args(["get-access-token", "--token-cache", choice])
            self.assertEqual(args.token_cache, choice)


class CliMainTest(unittest.TestCase):
    def _make_provider(self, access_token):
        provider = mock.MagicMock()
        provider.ensure_token.return_value = {"access_token": access_token}
        return provider

    @mock.patch("kognic.auth.cli.get_access_token.make_token_provider")
    def test_main_prints_token(self, mock_make_provider):
        mock_make_provider.return_value = self._make_provider("test-access-token-123")

        with mock.patch("builtins.print") as mock_print:
            result = main(["get-access-token", "--token-cache", "none"])

        self.assertEqual(result, 0)
        mock_print.assert_called_once_with("test-access-token-123")

    @mock.patch("kognic.auth.cli.get_access_token.make_token_provider")
    def test_main_with_credentials_file(self, mock_make_provider):
        mock_make_provider.return_value = self._make_provider("token-from-file")

        with mock.patch("builtins.print") as mock_print:
            result = main(["get-access-token", "--credentials", "/path/to/creds.json", "--token-cache", "none"])

        self.assertEqual(result, 0)
        mock_print.assert_called_once_with("token-from-file")
        mock_make_provider.assert_called_once_with(
            auth="/path/to/creds.json",
            auth_host=DEFAULT_HOST,
            token_cache=None,
        )

    @mock.patch("kognic.auth.cli.get_access_token.make_token_provider")
    def test_main_with_custom_server(self, mock_make_provider):
        mock_make_provider.return_value = self._make_provider("custom-server-token")

        with mock.patch("builtins.print") as mock_print:
            result = main(["get-access-token", "--server", "https://custom.server", "--token-cache", "none"])

        self.assertEqual(result, 0)
        mock_print.assert_called_once_with("custom-server-token")
        mock_make_provider.assert_called_once_with(
            auth=None,
            auth_host="https://custom.server",
            token_cache=None,
        )

    @mock.patch("kognic.auth.cli.get_access_token.make_token_provider")
    def test_main_with_all_options(self, mock_make_provider):
        mock_make_provider.return_value = self._make_provider("full-options-token")

        with mock.patch("builtins.print"):
            result = main(
                [
                    "get-access-token",
                    "--server",
                    "https://my.server",
                    "--credentials",
                    "creds.json",
                    "--token-cache",
                    "none",
                ]
            )

        self.assertEqual(result, 0)
        mock_make_provider.assert_called_once_with(
            auth="creds.json",
            auth_host="https://my.server",
            token_cache=None,
        )

    @mock.patch("kognic.auth.cli.get_access_token.make_token_provider")
    def test_main_file_not_found(self, mock_make_provider):
        mock_make_provider.side_effect = FileNotFoundError("Could not find Api Credentials file at /bad/path.json")

        with mock.patch("builtins.print") as mock_print:
            result = main(["get-access-token", "--credentials", "/bad/path.json", "--token-cache", "none"])

        self.assertEqual(result, 1)
        self.assertIn("Error:", mock_print.call_args[0][0])

    @mock.patch("kognic.auth.cli.get_access_token.make_token_provider")
    def test_main_value_error(self, mock_make_provider):
        mock_make_provider.side_effect = ValueError("Bad auth credentials")

        with mock.patch("builtins.print") as mock_print:
            result = main(["get-access-token", "--token-cache", "none"])

        self.assertEqual(result, 1)
        self.assertIn("Error:", mock_print.call_args[0][0])

    @mock.patch("kognic.auth.cli.get_access_token.make_token_provider")
    def test_main_generic_exception(self, mock_make_provider):
        mock_make_provider.side_effect = Exception("Network error")

        with mock.patch("builtins.print") as mock_print:
            result = main(["get-access-token", "--token-cache", "none"])

        self.assertEqual(result, 1)
        self.assertIn("Error fetching token:", mock_print.call_args[0][0])

    @mock.patch("kognic.auth.cli.get_access_token.make_token_provider")
    @mock.patch("kognic.auth.cli.get_access_token.load_kognic_env_config")
    def test_main_with_context(self, mock_load_config, mock_make_provider):
        from kognic.auth.env_config import KognicEnvConfig

        mock_load_config.return_value = KognicEnvConfig(
            environments={
                "demo": Environment(
                    name="demo",
                    host="demo.kognic.com",
                    auth_server="https://auth.demo.kognic.com",
                    credentials="/path/to/demo-creds.json",
                ),
            },
        )
        mock_make_provider.return_value = self._make_provider("demo-token")

        with mock.patch("builtins.print") as mock_print:
            result = main(["get-access-token", "--env", "demo", "--token-cache", "none"])

        self.assertEqual(result, 0)
        mock_print.assert_called_once_with("demo-token")
        mock_make_provider.assert_called_once_with(
            auth="/path/to/demo-creds.json",
            auth_host="https://auth.demo.kognic.com",
            token_cache=None,
        )

    @mock.patch("kognic.auth.cli.get_access_token.make_token_provider")
    @mock.patch("kognic.auth.cli.get_access_token.load_kognic_env_config")
    def test_main_with_context_server_override(self, mock_load_config, mock_make_provider):
        from kognic.auth.env_config import KognicEnvConfig

        mock_load_config.return_value = KognicEnvConfig(
            environments={
                "demo": Environment(
                    name="demo",
                    host="demo.kognic.com",
                    auth_server="https://auth.demo.kognic.com",
                    credentials="/path/to/demo-creds.json",
                ),
            },
        )
        mock_make_provider.return_value = self._make_provider("override-token")

        with mock.patch("builtins.print"):
            result = main(
                ["get-access-token", "--env", "demo", "--server", "https://custom.server", "--token-cache", "none"]
            )

        self.assertEqual(result, 0)
        mock_make_provider.assert_called_once_with(
            auth="/path/to/demo-creds.json",
            auth_host="https://custom.server",
            token_cache=None,
        )

    def test_main_with_unknown_context(self):
        with mock.patch("kognic.auth.cli.get_access_token.load_kognic_env_config") as mock_load_config:
            from kognic.auth.env_config import KognicEnvConfig

            mock_load_config.return_value = KognicEnvConfig()

            with mock.patch("builtins.print") as mock_print:
                result = main(["get-access-token", "--env", "nonexistent"])

        self.assertEqual(result, 1)
        self.assertIn("nonexistent", mock_print.call_args[0][0])


class CliCacheTest(unittest.TestCase):
    """Tests for token caching in get-access-token."""

    def _make_token(self, access_token="cached-token-abc"):
        return {
            "access_token": access_token,
            "expires_at": time.time() + 3600,
            "expires_in": 3600,
            "token_type": "bearer",
        }

    @mock.patch("kognic.auth.cli.get_access_token.make_token_provider")
    @mock.patch("kognic.auth.cli.get_access_token.make_cache")
    def test_cache_hit_injects_token_into_provider(self, mock_make_cache, mock_make_provider):
        """make_cache is called with 'auto' and its result is passed to make_token_provider."""
        mock_cache = mock.MagicMock()
        mock_make_cache.return_value = mock_cache

        provider = mock.MagicMock()
        provider.ensure_token.return_value = {"access_token": "cached-token-abc"}
        mock_make_provider.return_value = provider

        with mock.patch("builtins.print") as mock_print:
            result = main(["get-access-token"])

        self.assertEqual(result, 0)
        mock_make_cache.assert_called_once_with("auto")
        mock_make_provider.assert_called_once_with(
            auth=mock.ANY,
            auth_host=mock.ANY,
            token_cache=mock_cache,
        )
        mock_print.assert_called_once_with("cached-token-abc")

    @mock.patch("kognic.auth.cli.get_access_token.make_token_provider")
    @mock.patch("kognic.auth.cli.get_access_token.make_cache")
    def test_no_cache_passes_none_to_provider(self, mock_make_cache, mock_make_provider):
        """--token-cache none results in make_cache returning None and provider receiving None."""
        mock_make_cache.return_value = None

        provider = mock.MagicMock()
        provider.ensure_token.return_value = {"access_token": "fresh-token"}
        mock_make_provider.return_value = provider

        with mock.patch("builtins.print"):
            result = main(["get-access-token", "--token-cache", "none"])

        self.assertEqual(result, 0)
        mock_make_cache.assert_called_once_with("none")
        mock_make_provider.assert_called_once_with(
            auth=mock.ANY,
            auth_host=mock.ANY,
            token_cache=None,
        )

    @mock.patch("kognic.auth.cli.get_access_token.make_token_provider")
    @mock.patch("kognic.auth.cli.get_access_token.make_cache")
    def test_cache_mode_forwarded(self, mock_make_cache, mock_make_provider):
        """--token-cache keyring passes 'keyring' to make_cache."""
        mock_make_cache.return_value = mock.MagicMock()
        provider = mock.MagicMock()
        provider.ensure_token.return_value = {"access_token": "t"}
        mock_make_provider.return_value = provider

        with mock.patch("builtins.print"):
            main(["get-access-token", "--token-cache", "keyring"])

        mock_make_cache.assert_called_once_with("keyring")


class KogParserTest(unittest.TestCase):
    def test_kog_basic(self):
        parser = create_kog_parser()
        args = parser.parse_args(["get", "https://app.kognic.com/v1/projects"])
        self.assertEqual(args.method, "get")
        self.assertEqual(args.url, "https://app.kognic.com/v1/projects")
        self.assertIsNone(args.data)
        self.assertIsNone(args.headers)
        self.assertIsNone(args.env_name)

    def test_kog_with_method(self):
        parser = create_kog_parser()
        args = parser.parse_args(["post", "https://app.kognic.com/v1/projects"])
        self.assertEqual(args.method, "post")

    def test_kog_with_data(self):
        parser = create_kog_parser()
        args = parser.parse_args(["post", "https://app.kognic.com/v1/projects", "-d", '{"name": "test"}'])
        self.assertEqual(args.method, "post")
        self.assertEqual(args.data, '{"name": "test"}')

    def test_kog_with_headers(self):
        parser = create_kog_parser()
        args = parser.parse_args(
            [
                "get",
                "https://app.kognic.com/v1/projects",
                "-H",
                "Accept: application/json",
                "-H",
                "X-Custom: value",
            ]
        )
        self.assertEqual(args.headers, ["Accept: application/json", "X-Custom: value"])

    def test_kog_with_env(self):
        parser = create_kog_parser()
        args = parser.parse_args(["get", "https://demo.kognic.com/v1/projects", "--env", "demo"])
        self.assertEqual(args.env_name, "demo")

    def test_kog_with_config(self):
        parser = create_kog_parser()
        args = parser.parse_args(
            ["get", "https://app.kognic.com/v1/projects", "--env-config-file-path", "/custom/config.json"]
        )
        self.assertEqual(args.env_config_file_path, "/custom/config.json")

    def test_kog_token_cache_default(self):
        parser = create_kog_parser()
        args = parser.parse_args(["get", "https://app.kognic.com/v1/projects"])
        self.assertEqual(args.token_cache, "auto")

    def test_kog_token_cache_none(self):
        parser = create_kog_parser()
        args = parser.parse_args(["get", "https://app.kognic.com/v1/projects", "--token-cache", "none"])
        self.assertEqual(args.token_cache, "none")


class CallApiTest(unittest.TestCase):
    def _make_parsed(
        self,
        method="get",
        url="https://app.kognic.com/v1/projects",
        data=None,
        headers=None,
        env_config_file_path="/nonexistent/config.json",
        env_name=None,
        token_cache="none",
    ):
        parser = create_kog_parser()
        args = [method, url]
        if data:
            args.extend(["-d", data])
        if headers:
            for h in headers:
                args.extend(["-H", h])
        args.extend(["--env-config-file-path", env_config_file_path])
        if env_name:
            args.extend(["--env", env_name])
        args.extend(["--token-cache", token_cache])
        return parser.parse_args(args)

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_get_success(self, mock_create_session, mock_load_config, mock_resolve_environment):
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "application/json"}
        mock_response.json.return_value = {"projects": []}
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed()
        with mock.patch("builtins.print") as mock_print:
            result = call_run(parsed)

        self.assertEqual(result, 0)
        mock_session.request.assert_called_once_with(
            method="GET",
            url="https://app.kognic.com/v1/projects",
            json=None,
            headers=None,
        )
        mock_print.assert_called_once_with(json.dumps({"projects": []}, indent=2))

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_post_with_data(self, mock_create_session, mock_load_config, mock_resolve_environment):
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "application/json"}
        mock_response.json.return_value = {"id": 1}
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed(method="post", data='{"name": "test"}')
        with mock.patch("builtins.print"):
            result = call_run(parsed)

        self.assertEqual(result, 0)
        mock_session.request.assert_called_once_with(
            method="POST",
            url="https://app.kognic.com/v1/projects",
            json={"name": "test"},
            headers={"Content-Type": "application/json"},
        )

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_with_custom_headers(self, mock_create_session, mock_load_config, mock_resolve_environment):
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "text/plain"}
        mock_response.text = "OK"
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed(headers=["Accept: text/plain", "X-Custom: value"])
        with mock.patch("builtins.print"):
            result = call_run(parsed)

        self.assertEqual(result, 0)
        mock_session.request.assert_called_once_with(
            method="GET",
            url="https://app.kognic.com/v1/projects",
            json=None,
            headers={"Accept": "text/plain", "X-Custom": "value"},
        )

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_error_status(self, mock_create_session, mock_load_config, mock_resolve_environment):
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = False
        mock_response.headers = {"Content-Type": "application/json"}
        mock_response.json.return_value = {"error": "not found"}
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed()
        with mock.patch("builtins.print"):
            result = call_run(parsed)

        self.assertEqual(result, 1)

    def test_call_api_invalid_json_data(self):
        parsed = self._make_parsed(data="not json")
        with mock.patch("builtins.print") as mock_print:
            result = call_run(parsed)

        self.assertEqual(result, 1)
        self.assertIn("Invalid JSON data", mock_print.call_args[0][0])

    def test_call_api_invalid_header_format(self):
        parsed = self._make_parsed(headers=["BadHeader"])
        with mock.patch("builtins.print") as mock_print:
            result = call_run(parsed)

        self.assertEqual(result, 1)
        self.assertIn("Invalid header format", mock_print.call_args[0][0])

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_plain_text_response(self, mock_create_session, mock_load_config, mock_resolve_environment):
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "text/plain"}
        mock_response.text = "Hello World"
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed()
        with mock.patch("builtins.print") as mock_print:
            result = call_run(parsed)

        self.assertEqual(result, 0)
        mock_print.assert_called_once_with("Hello World")

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_jsonl_data_array(self, mock_create_session, mock_load_config, mock_resolve_environment):
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "application/json"}
        mock_response.json.return_value = {"data": [{"id": 1, "name": "a"}, {"id": 2, "name": "b"}]}
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed()
        parsed.output_format = "jsonl"
        with mock.patch("builtins.print") as mock_print:
            result = call_run(parsed)

        self.assertEqual(result, 0)
        self.assertEqual(mock_print.call_count, 2)
        mock_print.assert_any_call(json.dumps({"id": 1, "name": "a"}))
        mock_print.assert_any_call(json.dumps({"id": 2, "name": "b"}))

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_jsonl_single_key_non_data(self, mock_create_session, mock_load_config, mock_resolve_environment):
        """When --format=jsonl is used and response has a single key holding a list, flatten it."""
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "application/json"}
        mock_response.json.return_value = {"projects": [{"id": 1}, {"id": 2}]}
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed()
        parsed.output_format = "jsonl"
        with mock.patch("builtins.print") as mock_print:
            result = call_run(parsed)

        self.assertEqual(result, 0)
        self.assertEqual(mock_print.call_count, 2)
        mock_print.assert_any_call(json.dumps({"id": 1}))
        mock_print.assert_any_call(json.dumps({"id": 2}))

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_jsonl_multiple_keys(self, mock_create_session, mock_load_config, mock_resolve_environment):
        """When --format=jsonl is used but response has multiple keys, pretty-print as usual."""
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "application/json"}
        mock_response.json.return_value = {"data": [{"id": 1}], "total": 1}
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed()
        parsed.output_format = "jsonl"
        with mock.patch("builtins.print") as mock_print:
            result = call_run(parsed)

        self.assertEqual(result, 0)
        mock_print.assert_called_once_with(json.dumps({"data": [{"id": 1}], "total": 1}, indent=2))

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_jsonl_top_level_list(self, mock_create_session, mock_load_config, mock_resolve_environment):
        """When --format=jsonl is used and response body is a list, flatten it."""
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "application/json"}
        mock_response.json.return_value = [{"id": 1}, {"id": 2}, {"id": 3}]
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed()
        parsed.output_format = "jsonl"
        with mock.patch("builtins.print") as mock_print:
            result = call_run(parsed)

        self.assertEqual(result, 0)
        self.assertEqual(mock_print.call_count, 3)

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_jsonl_empty_data(self, mock_create_session, mock_load_config, mock_resolve_environment):
        """When --format=jsonl is used and data is an empty list, nothing is printed."""
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "application/json"}
        mock_response.json.return_value = {"data": []}
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed()
        parsed.output_format = "jsonl"
        with mock.patch("builtins.print") as mock_print:
            result = call_run(parsed)

        self.assertEqual(result, 0)
        mock_print.assert_not_called()

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_csv_data_array(self, mock_create_session, mock_load_config, mock_resolve_environment):
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "application/json"}
        mock_response.json.return_value = {"data": [{"id": 1, "name": "a"}, {"id": 2, "name": "b"}]}
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed()
        parsed.output_format = "csv"
        with mock.patch("builtins.print") as mock_print:
            result = call_run(parsed)

        self.assertEqual(result, 0)
        output = mock_print.call_args[0][0]
        lines = output.strip().split("\r\n")
        self.assertEqual(lines[0], "id,name")
        self.assertEqual(lines[1], "1,a")
        self.assertEqual(lines[2], "2,b")

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_tsv_data_array(self, mock_create_session, mock_load_config, mock_resolve_environment):
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "application/json"}
        mock_response.json.return_value = {"data": [{"id": 1, "name": "a"}, {"id": 2, "name": "b"}]}
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed()
        parsed.output_format = "tsv"
        with mock.patch("builtins.print") as mock_print:
            result = call_run(parsed)

        self.assertEqual(result, 0)
        output = mock_print.call_args[0][0]
        lines = output.strip().split("\r\n")
        self.assertEqual(lines[0], "id\tname")
        self.assertEqual(lines[1], "1\ta")
        self.assertEqual(lines[2], "2\tb")

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_table_data_array(self, mock_create_session, mock_load_config, mock_resolve_environment):
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "application/json"}
        mock_response.json.return_value = {"data": [{"id": 1, "name": "alice"}, {"id": 2, "name": "b"}]}
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed()
        parsed.output_format = "table"
        with mock.patch("builtins.print") as mock_print:
            result = call_run(parsed)

        self.assertEqual(result, 0)
        lines = [call[0][0] for call in mock_print.call_args_list]
        self.assertEqual(lines[0], "| id | name  |")
        self.assertEqual(lines[1], "|----|-------|")
        self.assertEqual(lines[2], "| 1  | alice |")
        self.assertEqual(lines[3], "| 2  | b     |")

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_table_empty_data(self, mock_create_session, mock_load_config, mock_resolve_environment):
        """Table with empty list prints nothing."""
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "application/json"}
        mock_response.json.return_value = {"data": []}
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed()
        parsed.output_format = "table"
        with mock.patch("builtins.print") as mock_print:
            result = call_run(parsed)

        self.assertEqual(result, 0)
        mock_print.assert_not_called()

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_csv_nested_values(self, mock_create_session, mock_load_config, mock_resolve_environment):
        """Nested dicts and lists are JSON-serialized in CSV output."""
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "application/json"}
        mock_response.json.return_value = {"data": [{"id": 1, "tags": ["a", "b"], "meta": {"key": "val"}}]}
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed()
        parsed.output_format = "csv"
        with mock.patch("builtins.print") as mock_print:
            result = call_run(parsed)

        self.assertEqual(result, 0)
        output = mock_print.call_args[0][0]
        lines = output.strip().split("\r\n")
        self.assertEqual(lines[0], "id,tags,meta")
        self.assertEqual(lines[1], '1,"[""a"", ""b""]","{""key"": ""val""}"')

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_table_nested_values(self, mock_create_session, mock_load_config, mock_resolve_environment):
        """Nested dicts and lists are JSON-serialized in table output."""
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "application/json"}
        mock_response.json.return_value = {"data": [{"id": 1, "tags": ["a", "b"]}]}
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed()
        parsed.output_format = "table"
        with mock.patch("builtins.print") as mock_print:
            result = call_run(parsed)

        self.assertEqual(result, 0)
        lines = [call[0][0] for call in mock_print.call_args_list]
        self.assertEqual(lines[0], "| id | tags       |")
        self.assertEqual(lines[1], "|----|------------|")
        self.assertEqual(lines[2], '| 1  | ["a", "b"] |')

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_csv_top_level_list(self, mock_create_session, mock_load_config, mock_resolve_environment):
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "application/json"}
        mock_response.json.return_value = [{"x": 10, "y": 20}, {"x": 30, "y": 40}]
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed()
        parsed.output_format = "csv"
        with mock.patch("builtins.print") as mock_print:
            result = call_run(parsed)

        self.assertEqual(result, 0)
        output = mock_print.call_args[0][0]
        lines = output.strip().split("\r\n")
        self.assertEqual(lines[0], "x,y")
        self.assertEqual(lines[1], "10,20")
        self.assertEqual(lines[2], "30,40")

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_csv_sparse_keys(self, mock_create_session, mock_load_config, mock_resolve_environment):
        """CSV output includes all keys across all rows, with blanks for missing values."""
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "application/json"}
        mock_response.json.return_value = {"data": [{"id": 1, "name": "a"}, {"id": 2, "extra": "z"}]}
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed()
        parsed.output_format = "csv"
        with mock.patch("builtins.print") as mock_print:
            result = call_run(parsed)

        self.assertEqual(result, 0)
        output = mock_print.call_args[0][0]
        lines = output.strip().split("\r\n")
        self.assertEqual(lines[0], "id,name,extra")
        self.assertEqual(lines[1], "1,a,")
        self.assertEqual(lines[2], "2,,z")

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_csv_empty_data(self, mock_create_session, mock_load_config, mock_resolve_environment):
        """CSV with empty list prints nothing."""
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "application/json"}
        mock_response.json.return_value = {"data": []}
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed()
        parsed.output_format = "csv"
        with mock.patch("builtins.print") as mock_print:
            result = call_run(parsed)

        self.assertEqual(result, 0)
        mock_print.assert_not_called()

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    def test_call_api_csv_not_flattenable(self, mock_create_session, mock_load_config, mock_resolve_environment):
        """CSV with non-flattenable response falls back to pretty JSON."""
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "application/json"}
        mock_response.json.return_value = {"a": 1, "b": 2}
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parsed = self._make_parsed()
        parsed.output_format = "csv"
        with mock.patch("builtins.print") as mock_print:
            result = call_run(parsed)

        self.assertEqual(result, 0)
        mock_print.assert_called_once_with(json.dumps({"a": 1, "b": 2}, indent=2))

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    def test_call_api_uses_context_credentials(self, mock_load_config, mock_resolve_environment):
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="demo",
            host="demo.kognic.com",
            auth_server="https://auth.demo.kognic.com",
            credentials="/path/to/demo-creds.json",
        )

        with (
            mock.patch("kognic.auth.cli.api_request.make_token_provider") as mock_make_provider,
            mock.patch("kognic.auth.cli.api_request.create_session") as mock_create_session,
        ):
            mock_session = mock.MagicMock()
            mock_response = mock.MagicMock()
            mock_response.ok = True
            mock_response.headers = {"Content-Type": "text/plain"}
            mock_response.text = "ok"
            mock_session.request.return_value = mock_response
            mock_create_session.return_value = mock_session

            parsed = self._make_parsed(url="https://demo.kognic.com/v1/projects")
            with mock.patch("builtins.print"):
                call_run(parsed)

            mock_make_provider.assert_called_once_with(
                auth="/path/to/demo-creds.json",
                auth_host="https://auth.demo.kognic.com",
                token_cache=None,
            )


class KogCacheTest(unittest.TestCase):
    """Tests for token caching in kog command."""

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.create_session")
    @mock.patch("kognic.auth.cli.api_request.make_token_provider")
    def test_kog_token_cache_none_forwarded(
        self, mock_make_provider, mock_create_session, mock_load_config, mock_resolve_environment
    ):
        """--token-cache none is forwarded as token_cache=None to make_token_provider."""
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials=None,
        )
        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "text/plain"}
        mock_response.text = "ok"
        mock_session.request.return_value = mock_response
        mock_create_session.return_value = mock_session

        parser = create_kog_parser()
        parsed = parser.parse_args(
            [
                "get",
                "https://app.kognic.com/v1/projects",
                "--env-config-file-path",
                "/nonexistent/config.json",
                "--token-cache",
                "none",
            ]
        )
        with mock.patch("builtins.print"):
            call_run(parsed)

        mock_make_provider.assert_called_once_with(
            auth=None,
            auth_host="https://auth.app.kognic.com",
            token_cache=None,
        )

    @mock.patch("kognic.auth.cli.api_request.resolve_environment")
    @mock.patch("kognic.auth.cli.api_request.load_kognic_env_config")
    @mock.patch("kognic.auth.cli.api_request.make_token_provider")
    @mock.patch("kognic.auth.cli.api_request.make_cache")
    def test_kog_uses_cached_token(
        self, mock_make_cache, mock_make_provider, mock_load_config, mock_resolve_environment
    ):
        """When a cached token is available, it is injected via make_token_provider."""
        mock_load_config.return_value = mock.MagicMock()
        mock_resolve_environment.return_value = Environment(
            name="default",
            host="app.kognic.com",
            auth_server="https://auth.app.kognic.com",
            credentials="/path/to/creds.json",
        )

        mock_cache = mock.MagicMock()
        mock_make_cache.return_value = mock_cache

        mock_provider = mock.MagicMock()
        mock_make_provider.return_value = mock_provider

        mock_session = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.headers = {"Content-Type": "application/json"}
        mock_response.json.return_value = {"result": "ok"}
        mock_session.request.return_value = mock_response

        with mock.patch("kognic.auth.cli.api_request.create_session", return_value=mock_session):
            parser = create_kog_parser()
            parsed = parser.parse_args(
                ["get", "https://app.kognic.com/v1/projects", "--env-config-file-path", "/nonexistent/config.json"]
            )
            with mock.patch("builtins.print"):
                result = call_run(parsed)

        self.assertEqual(result, 0)
        mock_make_cache.assert_called_once_with("auto")
        mock_make_provider.assert_called_once_with(
            auth="/path/to/creds.json",
            auth_host="https://auth.app.kognic.com",
            token_cache=mock_cache,
        )


class CredentialsCommandTest(unittest.TestCase):
    def test_put_stores_credentials(self):
        import json
        import tempfile

        creds = {
            "clientId": "test-client-id",
            "clientSecret": "test-secret",
            "email": "test@kognic.com",
            "userId": 1,
            "issuer": "auth.kognic.test",
        }
        with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as f:
            json.dump(creds, f)
            path = f.name

        try:
            with mock.patch("kognic.auth.cli.credentials.save_credentials") as mock_save:
                result = main(["credentials", "put", path])
            self.assertEqual(result, 0)
            args, kwargs = mock_save.call_args
            self.assertEqual(args[0].client_id, "test-client-id")
            self.assertEqual(args[0].client_secret, "test-secret")
            self.assertEqual(args[1], "default")
        finally:
            Path(path).unlink()

    def test_put_custom_profile(self):
        import json
        import tempfile

        creds = {
            "clientId": "id",
            "clientSecret": "secret",
            "email": "e@kognic.com",
            "userId": 1,
            "issuer": "issuer",
        }
        with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as f:
            json.dump(creds, f)
            path = f.name

        try:
            with mock.patch("kognic.auth.cli.credentials.save_credentials") as mock_save:
                result = main(["credentials", "put", path, "--env", "demo"])
            self.assertEqual(result, 0)
            args, kwargs = mock_save.call_args
            self.assertEqual(args[0].client_id, "id")
            self.assertEqual(args[0].client_secret, "secret")
            self.assertEqual(args[1], "demo")
        finally:
            Path(path).unlink()

    def test_put_missing_file_returns_error(self):
        result = main(["credentials", "put", "/nonexistent/creds.json"])
        self.assertEqual(result, 1)

    def test_get_returns_credentials(self):
        from kognic.auth.credentials_parser import ApiCredentials

        fake_creds = ApiCredentials(
            client_id="my-id",
            client_secret="my-secret",
            email="user@kognic.com",
            user_id=42,
            issuer="auth.kognic.com",
        )
        with mock.patch("kognic.auth.cli.credentials.load_credentials", return_value=fake_creds):
            with mock.patch("builtins.print") as mock_print:
                result = main(["credentials", "get"])
        self.assertEqual(result, 0)
        output = json.loads(mock_print.call_args[0][0])
        self.assertEqual(output["clientId"], "my-id")
        self.assertEqual(output["clientSecret"], "my-secret")
        self.assertEqual(output["email"], "user@kognic.com")
        self.assertEqual(output["userId"], 42)
        self.assertEqual(output["issuer"], "auth.kognic.com")

    def test_get_custom_profile(self):
        from kognic.auth.credentials_parser import ApiCredentials

        fake_creds = ApiCredentials(
            client_id="demo-id",
            client_secret="demo-secret",
            email="demo@kognic.com",
            user_id=1,
            issuer="auth.kognic.com",
        )
        with mock.patch("kognic.auth.cli.credentials.load_credentials", return_value=fake_creds) as mock_load:
            result = main(["credentials", "get", "--env", "demo"])
        self.assertEqual(result, 0)
        mock_load.assert_called_once_with("demo")

    def test_get_not_found_returns_error(self):
        with mock.patch("kognic.auth.cli.credentials.load_credentials", return_value=None):
            result = main(["credentials", "get"])
        self.assertEqual(result, 1)

    def test_delete_removes_credentials(self):
        with mock.patch("kognic.auth.cli.credentials.clear_credentials") as mock_clear:
            result = main(["credentials", "delete"])
        self.assertEqual(result, 0)
        mock_clear.assert_called_once_with("default")

    def test_delete_custom_profile(self):
        with mock.patch("kognic.auth.cli.credentials.clear_credentials") as mock_clear:
            result = main(["credentials", "delete", "--env", "demo"])
        self.assertEqual(result, 0)
        mock_clear.assert_called_once_with("demo")

    def test_no_subcommand_prints_help(self):
        with mock.patch("builtins.print"):
            result = main(["credentials"])
        self.assertEqual(result, 0)


if __name__ == "__main__":
    unittest.main()



"""Bearer token auth handler for requests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import requests
from requests.auth import AuthBase

if TYPE_CHECKING:
    from kognic.auth.requests.auth_session import RequestsAuthSession


class KognicBearerAuth(AuthBase):
    """Injects a Bearer token from a RequestsAuthSession into each request.

    Handles 401 responses by invalidating the cached token and retrying once.
    """

    def __init__(self, provider: RequestsAuthSession) -> None:
        self._provider = provider

    def __call__(self, r: requests.PreparedRequest) -> requests.PreparedRequest:
        r.headers["Authorization"] = f"Bearer {self._provider.ensure_token()['access_token']}"
        r.register_hook("response", self._handle_401)
        return r

    def _handle_401(self, r: requests.Response, **kwargs) -> requests.Response:
        if r.status_code != 401:
            return r
        self._provider.invalidate_token()
        _ = r.content  # drain socket so connection can be reused
        prep = r.request.copy()
        prep.headers["Authorization"] = f"Bearer {self._provider.ensure_token()['access_token']}"
        _r = r.connection.send(prep, **kwargs)
        _r.history.append(r)
        return _r



from __future__ import annotations

import json
import logging
from pathlib import Path
from typing import Optional

from kognic.auth import DEFAULT_CACHE_PATH
from kognic.auth.internal.token_cache._base import TokenCache, is_valid, make_key

log = logging.getLogger(__name__)


class FileTokenCache(TokenCache):
    """Token cache backed by a JSON file on disk."""

    def __init__(self, path: Path = DEFAULT_CACHE_PATH) -> None:
        self.path = path

    def _load_all(self) -> dict:
        try:
            return json.loads(self.path.read_text())
        except FileNotFoundError:
            return {}
        except Exception:
            log.debug("Failed to read token cache file", exc_info=True)
            return {}

    def _save_all(self, data: dict) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)
        self.path.write_text(json.dumps(data, indent=2))

    def load(self, auth_server: str, client_id: str) -> Optional[dict]:
        try:
            key = make_key(auth_server, client_id)
            token = self._load_all().get(key)
            if token is None:
                return None
            if not is_valid(token):
                log.debug("Cached file token expired or missing expires_at, discarding")
                return None
            log.debug("Using cached token from file (key=%s)", key)
            return token
        except Exception:
            log.debug("Failed to load token from file cache", exc_info=True)
            return None

    def save(self, auth_server: str, client_id: str, token: dict) -> None:
        try:
            key = make_key(auth_server, client_id)
            data = self._load_all()
            data[key] = token
            self._save_all(data)
            log.debug("Saved token to file cache for key=%s", key)
        except Exception:
            log.debug("Failed to save token to file cache", exc_info=True)

    def clear(self, auth_server: str, client_id: str) -> None:
        try:
            key = make_key(auth_server, client_id)
            data = self._load_all()
            if key in data:
                del data[key]
                self._save_all(data)
            log.debug("Cleared cached token from file cache for key=%s", key)
        except Exception:
            log.debug("Failed to clear token from file cache", exc_info=True)



import logging
import threading
from typing import Callable, Optional

import requests
from authlib.common.errors import AuthlibBaseError
from authlib.integrations.requests_client import OAuth2Session

from kognic.auth import DEFAULT_HOST, DEFAULT_TOKEN_ENDPOINT_RELPATH
from kognic.auth.base.auth_client import AuthClient
from kognic.auth.credentials_parser import resolve_credentials

log = logging.getLogger(__name__)


class _FixedSession(OAuth2Session):
    """OAuth2Session that gracefully recovers from an expired refresh token.

    Kognic's auth server issues a refresh token alongside the access token even
    for the client_credentials grant.  When the access token expires, authlib
    uses the refresh token to obtain a new one without a full re-authentication.
    If the refresh token has itself expired, plain OAuth2Session raises and
    propagates the error to every call site.

    This subclass intercepts the `invalid_token` error from a failed refresh and
    falls back to fetch_token() — a new client credentials grant — so the
    session self-heals transparently.
    """

    def refresh_token(self, url, **kwargs):
        try:
            super(_FixedSession, self).refresh_token(url, **kwargs)
        except AuthlibBaseError as e:
            if e.error == "invalid_token":
                log.info("Refresh token expired, resetting auth session")
                return self.fetch_token()
            raise
        except requests.exceptions.HTTPError as e:
            # with authlib >= 1.0.0
            if e.response.status_code == 401 and "invalid_token" == e.response.json().get("error"):
                log.info("Refresh token expired, resetting auth session")
                return self.fetch_token()
            raise


# https://docs.authlib.org/en/latest/client/oauth2.html
class RequestsAuthSession(AuthClient):
    """
    Not thread safe
    """

    def __init__(
        self,
        *,
        auth=None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        host: str = DEFAULT_HOST,
        token_endpoint: str = DEFAULT_TOKEN_ENDPOINT_RELPATH,
        initial_token: Optional[dict] = None,
        on_token_updated: Optional[Callable[[dict], None]] = None,
        **kwargs,
    ):
        """Initialize the auth session.

        Args:
            auth: Authentication credentials - path to credentials file or (client_id, client_secret) tuple
            client_id: Client id for authentication
            client_secret: Client secret for authentication
            host: Base url for authentication server
            token_endpoint: Relative path to the token endpoint
            initial_token: Pre-fetched token dict to inject, skipping the initial network fetch if valid.
            on_token_updated: Callback invoked with the new token dict whenever a fresh token is fetched.
            **kwargs: Additional params to pass into Client Constructor
        """
        self.host = host
        self.token_url = f"{host}{token_endpoint}"

        client_id, client_secret = resolve_credentials(auth, client_id, client_secret)
        self._client_id = client_id
        self._on_token_updated = on_token_updated

        self.oauth_session = _FixedSession(
            client_id=client_id,
            client_secret=client_secret,
            token_endpoint_auth_method="client_secret_post",
            grant_type="client_credentials",
            update_token=self._update_token,
            token_endpoint=self.token_url,
            token=initial_token,
            **kwargs,
        )
        self.oauth_session.register_compliance_hook("access_token_response", AuthClient.check_rate_limit)
        self.oauth_session.register_compliance_hook("refresh_token_response", AuthClient.check_rate_limit)

        self._lock = threading.RLock()

    @property
    def client_id(self) -> Optional[str]:
        return self._client_id

    @property
    def token(self):
        return self.oauth_session.token

    def _update_token(self, token, access_token=None, refresh_token=None):
        self._log_new_token()
        if self._on_token_updated is not None:
            self._on_token_updated(token)

    def ensure_token(self) -> dict:
        """Return a valid token, fetching one if needed. Thread-safe."""
        if not self.token:
            with self._lock:
                if not self.token:
                    token = self.oauth_session.fetch_access_token(url=self.token_url)
                    self._update_token(token)
        return self.token

    def invalidate_token(self) -> None:
        """Clear the cached token so the next ensure_token() call fetches a fresh one."""
        self.oauth_session.token = None

    @property
    def session(self):
        self.ensure_token()
        return self.oauth_session.session



from __future__ import annotations

import time
from abc import ABC, abstractmethod
from typing import Optional

SERVICE_NAME = "kognic-auth"

# Tokens are considered expired this many seconds before their actual expiry,
# to avoid using a token that expires mid-request.
EXPIRY_MARGIN_SECONDS = 30


def make_key(auth_server: str, client_id: str) -> str:
    return f"{auth_server}:{client_id}"


def is_valid(token: dict) -> bool:
    expires_at = token.get("expires_at")
    if expires_at is None:
        return False
    return time.time() < (expires_at - EXPIRY_MARGIN_SECONDS)


class TokenCache(ABC):
    """Abstract base class for token caches."""

    @abstractmethod
    def load(self, auth_server: str, client_id: str) -> Optional[dict]:
        """Return a non-expired token dict, or None."""

    @abstractmethod
    def save(self, auth_server: str, client_id: str, token: dict) -> None:
        """Persist a token dict. Silently ignores errors."""

    @abstractmethod
    def clear(self, auth_server: str, client_id: str) -> None:
        """Remove a cached token. Silently ignores errors."""



"""Protocol definitions for URL, Request, and Response types."""

from typing import Dict, Protocol, Union, runtime_checkable


@runtime_checkable
class Url(Protocol):
    """Protocol for URL objects (httpx URL)."""

    scheme: str
    host: str
    path: str


@runtime_checkable
class Request(Protocol):
    """Protocol for HTTP request objects."""

    method: str
    url: Union[str, Url]


@runtime_checkable
class Response(Protocol):
    """Protocol for HTTP response objects."""

    headers: Dict[str, str]
    request: Request

    def json(self) -> dict:
        raise NotImplementedError



"""Keyring-based storage for Kognic API client credentials (full credentials file)."""

from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from kognic.auth.credentials_parser import ApiCredentials

SERVICE_NAME = "kognic-credentials"
DEFAULT_PROFILE = "default"

log = logging.getLogger(__name__)


def _get_keyring():
    """Return the keyring module if a usable backend is available, else None."""
    try:
        import keyring

        backend = keyring.get_keyring()
        if "fail" in type(backend).__name__.lower():
            return None
        return keyring
    except Exception:
        return None


def load_credentials(profile: str = DEFAULT_PROFILE) -> Optional[ApiCredentials]:
    """Load full credentials from keyring, or None if not found."""
    kr = _get_keyring()
    if kr is None:
        return None
    try:
        stored = kr.get_password(SERVICE_NAME, profile)
        if stored is None:
            return None

        from kognic.auth.credentials_parser import parse_credentials

        return parse_credentials(json.loads(stored))
    except Exception:
        log.debug("Failed to load credentials from keyring", exc_info=True)
        return None


def save_credentials(creds: ApiCredentials, profile: str = DEFAULT_PROFILE) -> None:
    """Store full credentials in keyring. Raises RuntimeError if keyring is unavailable."""
    kr = _get_keyring()
    if kr is None:
        raise RuntimeError(
            "No usable keyring backend available. "
            "Install a keyring backend (e.g. 'pip install keyring') or use environment variables instead."
        )
    data = {
        "clientId": creds.client_id,
        "clientSecret": creds.client_secret,
        "email": creds.email,
        "userId": creds.user_id,
        "issuer": creds.issuer,
    }
    kr.set_password(SERVICE_NAME, profile, json.dumps(data))
    log.debug("Saved credentials to keyring for profile=%s", profile)


def clear_credentials(profile: str = DEFAULT_PROFILE) -> None:
    """Remove credentials from keyring. Silently does nothing if not found or keyring unavailable."""
    kr = _get_keyring()
    if kr is None:
        return
    try:
        kr.delete_password(SERVICE_NAME, profile)
        log.debug("Cleared credentials from keyring for profile=%s", profile)
    except Exception:
        log.debug("Failed to clear credentials from keyring", exc_info=True)



"""Unit tests for BaseApiClient (sync client)."""

import json
import tempfile
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from kognic.auth._sunset import DATETIME_FMT, handle_sunset


class TestSunsetHeaderHandling(unittest.TestCase):
    def _make_mock_response(self, url: str, headers: dict, method: str = "GET"):
        response = MagicMock()
        response.request.url = url
        response.request.method = method
        response.headers = headers
        response.status_code = 200
        return response

    def test_no_sunset_header(self):
        response = self._make_mock_response("https://api.example.com/v1/test", {})
        # Should not log anything
        with patch("kognic.auth._sunset.logger") as mock_logger:
            handle_sunset(response)
            mock_logger.warning.assert_not_called()
            mock_logger.error.assert_not_called()

    def test_sunset_header_logs_warning_when_far(self):
        # Sunset date far in the future (> 2 weeks)
        from datetime import datetime, timedelta

        future_date = datetime.now() + timedelta(days=30)
        sunset_str = future_date.strftime(DATETIME_FMT)

        response = self._make_mock_response("https://api.example.com/v1/sunset-test", {"sunset-date": sunset_str})

        with patch("kognic.auth._sunset.logger") as mock_logger:
            handle_sunset(response)
            mock_logger.warning.assert_called_once()
            call_args = mock_logger.warning.call_args[0][0]
            self.assertIn("deprecated", call_args)
            self.assertIn("sunset-test", call_args)

    def test_sunset_header_logs_error_when_close(self):
        # Sunset date close (< 2 weeks)
        from datetime import datetime, timedelta

        close_date = datetime.now() + timedelta(days=7)
        sunset_str = close_date.strftime(DATETIME_FMT)

        response = self._make_mock_response("https://api.example.com/v1/sunset-test", {"sunset-date": sunset_str})

        with patch("kognic.auth._sunset.logger") as mock_logger:
            handle_sunset(response)
            mock_logger.error.assert_called_once()


class TestBaseApiClient(unittest.TestCase):
    @patch("kognic.auth.requests.base_client.RequestsAuthSession")
    def test_session_lazy_init(self, mock_session_class):
        from kognic.auth.requests.base_client import BaseApiClient

        mock_instance = MagicMock()
        mock_instance.session = MagicMock()
        mock_session_class.return_value = mock_instance

        client = BaseApiClient(auth=("test", "secret"))

        # Session should not be created yet
        self.assertIsNone(client._session)

        # Access session to trigger lazy init
        _ = client.session

        # Now session should be created
        mock_session_class.assert_called_once()

    @patch("kognic.auth.requests.base_client.RequestsAuthSession")
    def test_client_name_auto(self, mock_session_class):
        from kognic.auth.requests.base_client import BaseApiClient

        mock_instance = MagicMock()
        mock_instance.session = MagicMock()
        mock_session_class.return_value = mock_instance

        class MyCustomClient(BaseApiClient):
            pass

        client = MyCustomClient(auth=("test", "secret"))
        self.assertEqual(client._client_name, "MyCustomClient")


class TestBaseApiClientFromEnv(unittest.TestCase):
    def _write_config(self, data):
        f = tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False)
        json.dump(data, f)
        f.flush()
        f.close()
        return f.name

    @patch("kognic.auth.requests.base_client.RequestsAuthSession")
    def test_from_env_sets_auth_and_host(self, mock_session_class):
        from kognic.auth.requests.base_client import BaseApiClient

        mock_instance = MagicMock()
        mock_instance.session = MagicMock()
        mock_session_class.return_value = mock_instance

        config_path = self._write_config(
            {
                "environments": {
                    "demo": {
                        "host": "demo.kognic.com",
                        "auth_server": "https://auth.demo.kognic.com",
                        "credentials": "/tmp/demo-creds.json",
                    }
                }
            }
        )
        try:
            client = BaseApiClient.from_env("demo", env_config_path=config_path)
            self.assertEqual(client._auth, "/tmp/demo-creds.json")
            self.assertEqual(client._auth_host, "https://auth.demo.kognic.com")
        finally:
            Path(config_path).unlink()

    @patch("kognic.auth.requests.base_client.RequestsAuthSession")
    def test_explicit_auth_overrides_env_credentials(self, mock_session_class):
        from kognic.auth.requests.base_client import BaseApiClient

        mock_instance = MagicMock()
        mock_instance.session = MagicMock()
        mock_session_class.return_value = mock_instance

        config_path = self._write_config(
            {
                "environments": {
                    "demo": {
                        "host": "demo.kognic.com",
                        "auth_server": "https://auth.demo.kognic.com",
                        "credentials": "/tmp/demo-creds.json",
                    }
                }
            }
        )
        try:
            client = BaseApiClient.from_env("demo", env_config_path=config_path, auth=("my-id", "my-secret"))
            self.assertEqual(client._auth, ("my-id", "my-secret"))
            self.assertEqual(client._auth_host, "https://auth.demo.kognic.com")
        finally:
            Path(config_path).unlink()

    def test_unknown_env_raises(self):
        from kognic.auth.requests.base_client import BaseApiClient

        config_path = self._write_config({"environments": {}})
        try:
            with self.assertRaises(ValueError) as cm:
                BaseApiClient.from_env("nonexistent", env_config_path=config_path)
            self.assertIn("Unknown environment", str(cm.exception))
        finally:
            Path(config_path).unlink()

    @patch("kognic.auth.requests.base_client.RequestsAuthSession")
    def test_from_env_works_on_subclass(self, mock_session_class):
        from kognic.auth.requests.base_client import BaseApiClient

        mock_instance = MagicMock()
        mock_instance.session = MagicMock()
        mock_session_class.return_value = mock_instance

        class MyClient(BaseApiClient):
            pass

        config_path = self._write_config(
            {
                "environments": {
                    "demo": {
                        "host": "demo.kognic.com",
                        "auth_server": "https://auth.demo.kognic.com",
                        "credentials": "/tmp/demo-creds.json",
                    }
                }
            }
        )
        try:
            client = MyClient.from_env("demo", env_config_path=config_path)
            self.assertIsInstance(client, MyClient)
            self.assertEqual(client._client_name, "MyClient")
        finally:
            Path(config_path).unlink()


class TestProviderPool(unittest.TestCase):
    def setUp(self):
        import kognic.auth.requests.base_client as bc

        bc._provider_pool.clear()

    def tearDown(self):
        import kognic.auth.requests.base_client as bc

        bc._provider_pool.clear()

    def _make_clients(self, mock_session, n=2, **kwargs):
        from kognic.auth.requests.base_client import BaseApiClient

        clients = [BaseApiClient(**kwargs) for _ in range(n)]
        for c in clients:
            _ = c.session
        return clients

    @patch("kognic.auth.requests.base_client.requests.Session")
    @patch("kognic.auth.requests.base_client.RequestsAuthSession")
    @patch("kognic.auth.requests.base_client.resolve_credentials", return_value=("id1", "secret1"))
    def test_same_credentials_share_provider(self, _resolve, mock_ras, mock_session):
        self._make_clients(mock_session, n=2, auth=("id1", "secret1"))

        mock_ras.assert_called_once()
        self.assertEqual(len(mock_session.return_value.mount.call_args_list), 4)  # 2x per session

    @patch("kognic.auth.requests.base_client.requests.Session")
    @patch("kognic.auth.requests.base_client.RequestsAuthSession")
    @patch(
        "kognic.auth.requests.base_client.resolve_credentials",
        side_effect=lambda auth, *a, **kw: auth,
    )
    def test_different_credentials_get_different_providers(self, _resolve, mock_ras, mock_session):
        from kognic.auth.requests.base_client import BaseApiClient

        c1 = BaseApiClient(auth=("id1", "secret1"))
        c2 = BaseApiClient(auth=("id2", "secret2"))
        _ = c1.session
        _ = c2.session

        self.assertEqual(mock_ras.call_count, 2)

    @patch("kognic.auth.requests.base_client.requests.Session")
    @patch("kognic.auth.requests.base_client.RequestsAuthSession")
    @patch("kognic.auth.requests.base_client.resolve_credentials", return_value=("id1", "secret1"))
    def test_different_auth_host_gets_different_provider(self, _resolve, mock_ras, mock_session):
        from kognic.auth.requests.base_client import BaseApiClient

        c1 = BaseApiClient(auth=("id1", "secret1"), auth_host="https://auth.a.kognic.com")
        c2 = BaseApiClient(auth=("id1", "secret1"), auth_host="https://auth.b.kognic.com")
        _ = c1.session
        _ = c2.session

        self.assertEqual(mock_ras.call_count, 2)

    @patch("kognic.auth.requests.base_client.requests.Session")
    @patch("kognic.auth.requests.base_client.RequestsAuthSession")
    @patch("kognic.auth.requests.base_client.resolve_credentials", return_value=("id1", "secret1"))
    def test_cache_type_is_part_of_pool_key(self, _resolve, mock_ras, mock_session):
        from kognic.auth.internal.token_cache import FileTokenCache
        from kognic.auth.requests.base_client import BaseApiClient

        c1 = BaseApiClient(auth=("id1", "secret1"))
        c2 = BaseApiClient(auth=("id1", "secret1"), token_cache=FileTokenCache())
        _ = c1.session
        _ = c2.session

        self.assertEqual(mock_ras.call_count, 2)

    @patch("kognic.auth.requests.base_client.requests.Session")
    @patch("kognic.auth.requests.base_client.RequestsAuthSession")
    @patch("kognic.auth.requests.base_client.resolve_credentials", return_value=("id1", "secret1"))
    def test_explicit_token_provider_bypasses_pool(self, mock_resolve, mock_ras, mock_session):
        from kognic.auth.requests.base_client import BaseApiClient, _provider_pool

        explicit = MagicMock()
        client = BaseApiClient(auth=("id1", "secret1"), token_provider=explicit)
        _ = client.session

        self.assertEqual(len(_provider_pool), 0)
        mock_ras.assert_not_called()
        mock_resolve.assert_not_called()

    @patch("kognic.auth.requests.base_client.requests.Session")
    @patch("kognic.auth.requests.base_client.RequestsAuthSession")
    @patch("kognic.auth.requests.base_client.resolve_credentials", return_value=("id1", "secret1"))
    def test_pool_entry_alive_while_client_referenced(self, _resolve, mock_ras, mock_session):
        from kognic.auth.requests.base_client import (
            DEFAULT_HOST,
            DEFAULT_TOKEN_ENDPOINT_RELPATH,
            BaseApiClient,
            _provider_pool,
        )

        client = BaseApiClient(auth=("id1", "secret1"))
        _ = client.session

        pool_key = ("id1", DEFAULT_HOST, DEFAULT_TOKEN_ENDPOINT_RELPATH, type(None))
        self.assertIn(pool_key, _provider_pool)
        _ = client  # keep alive

    def test_provider_gc_when_all_clients_deleted(self):
        import gc
        import weakref

        from kognic.auth.requests.base_client import (
            DEFAULT_HOST,
            DEFAULT_TOKEN_ENDPOINT_RELPATH,
            BaseApiClient,
            _provider_pool,
        )

        # Use side_effect so each call returns a fresh object with no external strong references
        with patch(
            "kognic.auth.requests.base_client.resolve_credentials",
            return_value=("id-gc", "secret-gc"),
        ):
            with patch(
                "kognic.auth.requests.base_client.RequestsAuthSession",
                side_effect=lambda **kw: MagicMock(),
            ):
                client = BaseApiClient(auth=("id-gc", "secret-gc"))
                session = client.session
                pool_key = ("id-gc", DEFAULT_HOST, DEFAULT_TOKEN_ENDPOINT_RELPATH, type(None))
                provider_ref = weakref.ref(_provider_pool[pool_key])
                self.assertIsNotNone(provider_ref())

                del session, client
                gc.collect()
                gc.collect()  # two passes for cycles from _monkey_patch_send closures

                self.assertIsNone(provider_ref())


if __name__ == "__main__":
    unittest.main()
